import click

from au.click import AliasedGroup

from .eval_assignment import eval_assignment_cmd
from .gen_feedback import gen_feedback_cmd
from .gen_grades_csv import gen_grades_csv_cmd
from .quick_grade import quick_grade


@click.group(cls=AliasedGroup)
def python():
    """Commands for working with Python assignments."""


python.add_command(eval_assignment_cmd)
python.add_command(gen_feedback_cmd)
python.add_command(gen_grades_csv_cmd)
python.add_command(quick_grade)


if __name__ == "__main__":
    python()



import click
import functools
import logging


class DebugOptions:
    def __init__(self, base_level=logging.INFO, quiet_level=logging.WARNING):
        self.base_level = base_level
        self.quiet_level = quiet_level

    def options(self, func):
        @click.option(
            "-d", "--debug", is_flag=True, help="set to enable detailed output"
        )
        @click.option(
            "-q", "--quiet", is_flag=True, help="set to reduce output to errors only"
        )
        @functools.wraps(func)
        def command_wrapper(*args, **kwargs):
            if kwargs.get("debug"):
                logging.getLogger().setLevel(logging.DEBUG)
            elif kwargs.get("quiet"):
                logging.getLogger().setLevel(self.quiet_level)
            else:
                logging.getLogger().setLevel(self.base_level)
            return func(*args, **kwargs)

        return command_wrapper



from .AliasedGroup import AliasedGroup
from .SubdirGroup import SubdirGroup
from .AssignmentOptions import AssignmentOptions
from .RosterOptions import RosterOptions
from .DebugOptions import DebugOptions
from .BasePathType import BasePathType as BasePath




import logging
from pathlib import Path
import csv

import click

from au.click import BasePath, DebugOptions
from au.common.datetime import get_friendly_local_datetime

from .gen_feedback import get_feedback_file_score, DEFAULT_FEEDBACK_FILE_NAME
from .eval_assignment import retrieve_student_results


logger = logging.getLogger(__name__)


DEFAULT_GRADES_FILE_NAME = "grades.csv"


@click.command("gen-grades-csv")
@click.argument("root_dir", type=BasePath(), default=".")
@click.option(
    "--feedback-filename",
    type=str,
    default=DEFAULT_FEEDBACK_FILE_NAME,
    help="name of markdown file to generate",
)
@click.option(
    "--grades-filename",
    type=str,
    default=DEFAULT_GRADES_FILE_NAME,
    help="name of CSV file containing student feedback",
)
@click.option(
    "-y",
    "--skip-confirm",
    is_flag=True,
    help="set to bypass confirmation and overwrite existing CSV",
)
@DebugOptions().options
def gen_grades_csv_cmd(
    root_dir: Path,
    grades_filename: str = DEFAULT_GRADES_FILE_NAME,
    feedback_filename: str = DEFAULT_FEEDBACK_FILE_NAME,
    skip_confirm: bool = False,
    **kwargs,
) -> None:
    """Generate grades CSV file from feedback files in ROOT_DIR."""
    gen_grades_csv(root_dir, grades_filename, feedback_filename, skip_confirm)


def gen_grades_csv(
    root_dir: Path,
    grades_filename: str = DEFAULT_GRADES_FILE_NAME,
    feedback_filename: str = DEFAULT_FEEDBACK_FILE_NAME,
    skip_confirm: bool = False,
) -> None:
    """Generate grades CSV file from feedback files in ROOT_DIR."""

    grades_file = root_dir / grades_filename

    if not skip_confirm and grades_file.exists():
        click.confirm(f"Overwrite {grades_file}", abort=True)

    csv_header = [
        "Name",
        "Score",
        "Past Due",
        "Commit Count",
        "Last Commit",
        "Directory",
    ]
    csv_rows = []

    dirs = list(root_dir.iterdir())
    dirs.sort()

    for student_dir in dirs:
        if not student_dir.is_dir():
            continue

        try:
            student_results = retrieve_student_results(student_dir)
        except:
            logger.warning(f"No evaluation results found in {student_dir.name}")
            continue

        feedback_file = student_dir / feedback_filename
        if not feedback_file.exists():
            logger.warning(f"{feedback_filename} not found in {student_dir.name}")
            continue

        try:
            score = get_feedback_file_score(feedback_file)
        except:
            logger.error(f"Error retrieving score from {student_dir.name}")
            continue

        row = [student_results["name"], score]
        pd = student_results.setdefault("past_due", None)
        if pd:
            row.append(pd)
        else:
            row.append("")
        num_commits = student_results["num_commits"]
        row.append(num_commits)
        if num_commits:
            row.append(
                get_friendly_local_datetime(student_results["commit_date"]),
            )
        else:
            row.append("")
        row.append(student_results["dir_name"])

        csv_rows.append(row)

    with open(grades_file, "w") as csv_file:
        writer = csv.writer(csv_file)
        writer.writerow(csv_header)
        writer.writerows(csv_rows)

    print(f"Processed {len(csv_rows)} student grades")
    print(f"Generated {grades_filename} in {root_dir}")


if __name__ == "__main__":
    gen_grades_csv_cmd()



from pathlib import Path
from tomlkit import loads, dumps, TOMLDocument
from tomlkit.container import Container

class SettingsBase:
    FILENAME = 'settings.toml'

    def __init__(self, settings_doc_file: Path|str, create: bool = False):
        file = Path(settings_doc_file)
        exists = file.exists()
        if not create and not exists:
            raise FileNotFoundError(f"{file} does not exist")
        self._settings_doc_file = file
        self.is_dirty = False
        if exists:
            self.read()
        else:
            self.settings_doc = TOMLDocument()
            self.write()
            

    def __enter__(self):
        if not self.is_dirty:
            self.read()
        return self

    def __exit__(self, type, value, traceback):
        if self.is_dirty:
            self.write()

    def get(self, dotted_name: str, is_path=False) -> any:
        parts = dotted_name.split('.')
        doc = self.settings_doc
        for part in parts:
            doc = doc.get(part)
            if not doc:
                return None
        val = doc
        if is_path:
            path = Path(val)
            if not path.is_absolute():
                path = (self.settings_doc_path / path).resolve()
            val = path
        return val

    def set(self, dotted_name: str, val: any) -> any:
        parts = dotted_name.split('.')
        doc = self.settings_doc
        for part in parts[:-1]:
            doc = doc.setdefault(part, Container())
        if val is None:
            doc[parts[-1]] = ''
        elif isinstance(val, Path):
            try:
                val = val.resolve()
                doc_path = self.settings_doc_path.resolve()
                relative_path = val.relative_to(doc_path, walk_up=True).as_posix()
                doc[parts[-1]] = str(relative_path)
            except:
                doc[parts[-1]] = str(val)
        else:
            try:
                doc[parts[-1]] = val
            except:
                doc[parts[-1]] = str(val)
        self.is_dirty = True
    
    @property
    def settings_doc_path(self) -> Path:
        return self._settings_doc_file.parent

    @property
    def settings_doc(self) -> TOMLDocument:
        return self._settings_doc
    @settings_doc.setter
    def settings_doc(self, value):
        self._settings_doc = value

    def read(self):
        with open(self._settings_doc_file, 'rt') as fi:
            self.settings_doc = loads(fi.read())
            self.is_dirty = False

    def write(self):
        with open(self._settings_doc_file, 'wt') as fi:
            fi.write(dumps(self.settings_doc))
        self.is_dirty = False



from enum import Enum

DEFAULT_LINE_WIDTH = 100


class BoxChars(Enum):
    """Unicode box drawing characters."""

    LIGHT_HORIZONTAL = "\u2500"
    HEAVY_HORIZONTAL = "\u2501"
    LIGHT_VERTICAL = "\u2502"
    HEAVY_VERTICAL = "\u2503"
    LIGHT_DOWN_AND_RIGHT = "\u250c"
    HEAVY_DOWN_AND_RIGHT = "\u250f"
    LIGHT_DOWN_AND_LEFT = "\u2510"
    HEAVY_DOWN_AND_LEFT = "\u2513"
    LIGHT_UP_AND_RIGHT = "\u2514"
    HEAVY_UP_AND_RIGHT = "\u2517"
    LIGHT_UP_AND_LEFT = "\u2518"
    HEAVY_UP_AND_LEFT = "\u251b"
    LIGHT_VERTICAL_AND_RIGHT = "\u251c"
    HEAVY_VERTICAL_AND_RIGHT = "\u2523"
    LIGHT_VERTICAL_AND_LEFT = "\u2524"
    HEAVY_VERTICAL_AND_LEFT = "\u252b"
    LIGHT_DOWN_AND_HORIZONTAL = "\u252c"
    HEAVY_DOWN_AND_HORIZONTAL = "\u2533"
    LIGHT_UP_AND_HORIZONTAL = "\u2534"
    HEAVY_UP_AND_HORIZONTAL = "\u253b"
    LIGHT_VERTICAL_AND_HORIZONTAL = "\u253c"
    HEAVY_VERTICAL_AND_HORIZONTAL = "\u254b"
    LIGHT_DOUBLE_DASH_HORIZONTAL = "\u254c"
    HEAVY_DOUBLE_DASH_HORIZONTAL = "\u254d"
    LIGHT_DOUBLE_DASH_VERTICAL = "\u254e"
    HEAVY_DOUBLE_DASH_VERTICAL = "\u254f"
    DOUBLE_HORIZONTAL = "\u2550"
    DOUBLE_VERTICAL = "\u2551"
    DOUBLE_DOWN_AND_RIGHT = "\u2554"
    DOUBLE_DOWN_AND_LEFT = "\u2557"
    DOUBLE_UP_AND_RIGHT = "\u255a"
    DOUBLE_UP_AND_LEFT = "\u255d"
    DOUBLE_VERTICAL_AND_RIGHT = "\u2560"
    DOUBLE_VERTICAL_AND_LEFT = "\u2563"
    DOUBLE_DOWN_AND_HORIZONTAL = "\u2566"
    DOUBLE_UP_AND_HORIZONTAL = "\u2569"
    DOUBLE_VERTICAL_AND_HORIZONTAL = "\u256c"
    DOUBLE_VERTICAL_AND_HEAVY_RIGHT = "\u2561"
    DOUBLE_VERTICAL_AND_HEAVY_LEFT = "\u2562"
    DOUBLE_DOWN_AND_HEAVY_HORIZONTAL = "\u2564"
    DOUBLE_UP_AND_HEAVY_HORIZONTAL = "\u2567"
    DOUBLE_VERTICAL_AND_HEAVY_HORIZONTAL = "\u256a"
    LIGHT_VERTICAL_AND_DOUBLE_RIGHT = "\u255f"
    LIGHT_VERTICAL_AND_DOUBLE_LEFT = "\u2562"
    LIGHT_DOWN_AND_DOUBLE_HORIZONTAL = "\u2565"
    LIGHT_UP_AND_DOUBLE_HORIZONTAL = "\u2568"
    LIGHT_VERTICAL_AND_DOUBLE_HORIZONTAL = "\u256b"
    LIGHT_ARC_DOWN_AND_RIGHT = "\u256d"
    LIGHT_ARC_DOWN_AND_LEFT = "\u256e"
    LIGHT_ARC_UP_AND_LEFT = "\u256f"
    LIGHT_ARC_UP_AND_RIGHT = "\u2570"
    HEAVY_ARC_DOWN_AND_RIGHT = "\u2571"
    HEAVY_ARC_DOWN_AND_LEFT = "\u2572"
    HEAVY_ARC_UP_AND_LEFT = "\u2573"
    HEAVY_ARC_UP_AND_RIGHT = "\u2574"

    def __str__(self):
        return self.value


def get_single_line(text: str = "", width: int = DEFAULT_LINE_WIDTH) -> str:
    get_line(BoxChars.HEAVY_HORIZONTAL, text, width)


def get_double_line(text: str = "", width: int = DEFAULT_LINE_WIDTH) -> str:
    get_line(BoxChars.DOUBLE_HORIZONTAL, text, width)


def get_line(
    char: str = BoxChars.HEAVY_HORIZONTAL,
    text: str = "",
    width: int = DEFAULT_LINE_WIDTH,
) -> str:
    char = str(char)
    if text:
        text = text.strip()
        text_len = len(text) + 2
        if text_len >= width:
            return f"{char} {text}"
        else:
            line_len = width - text_len
            left_len = line_len // 2
            right_len = left_len + (line_len % 2)
            return f"{char * left_len} {text} {char * right_len}"
    else:
        return char * width



#!/usr/bin/env python3

from typing import List, Dict, overload

import subprocess
import json as json_module
from pprint import pformat

from dacite import from_dict

from .classroom_types import (
    Classroom,
    Assignment,
    AcceptedAssignment,
    _github_json_deserializer,
)
from au.common import draw_single_line, select_choice
from au.common.datetime import get_friendly_local_datetime, utc_min

import logging

logger = logging.getLogger(__name__)


def _gh_api(endpoint: str) -> Dict[str, any]:
    cmd = ["gh", "api", endpoint, "--paginate"]
    try:
        result = subprocess.run(cmd, capture_output=True, text=True, check=True)
        json = json_module.loads(result.stdout, object_hook=_github_json_deserializer)
        if isinstance(json, dict):
            if json.get("message", "").upper() == "NOT FOUND":
                logger.error(f"Invalid GitHub API Endpoint: {endpoint}")
                return None
        logger.debug(f"Retrieved JSON: {pformat(json)}")
        return json
    except (FileNotFoundError, subprocess.CalledProcessError):
        logger.exception(f"Error trying to run `gh` command")
        return None
    except json_module.JSONDecodeError as ex:
        logger.exception(f"Invalid JSON output from `gh` command")
        return None


###############################################################################
# get_classrooms / get_classroom / choose_classrooms
###############################################################################


def get_classrooms(include_archived=False) -> List[Classroom]:
    classrooms: List[Classroom] = []

    rooms_d = _gh_api("classrooms")
    if not rooms_d:
        return []
    for room_d in rooms_d:
        try:
            classroom = from_dict(Classroom, room_d)
        except:
            logger.exception("Error converting Classroom")
            return None
        classrooms.append(classroom)
    if not include_archived:
        classrooms = [c for c in classrooms if not c.archived]
    return classrooms


def get_classroom(classroom_id: int) -> Classroom:
    room_d = _gh_api(f"classrooms/{classroom_id}")
    if not room_d:
        return None
    try:
        classroom = from_dict(Classroom, room_d)
    except:
        logger.exception("Error converting Classroom")
        return None
    return classroom


def choose_classroom(
    include_archived: bool = False, suppress_print: bool = False
) -> Classroom:
    classrooms = get_classrooms(include_archived=include_archived)
    classrooms.reverse()
    choices = [c.name for c in classrooms]
    choice = select_choice(choices, title="CHOOSE CLASSROOM")
    if choice is not None:
        if not suppress_print:
            print(" > CHOICE:", choices[choice])
        return classrooms[choice]
    else:
        if not suppress_print:
            print(" X CHOICE: NONE")
        return None


###############################################################################
# get_assignments / choose_assignment / get_assignment
###############################################################################


@overload
def get_assignments(classroom: int) -> List[Assignment]: ...


@overload
def get_assignments(classroom: Classroom) -> List[Assignment]: ...


def get_assignments(classroom: Classroom | int) -> List[Assignment]:
    if isinstance(classroom, Classroom):
        classroom_id = classroom.id
    elif isinstance(classroom, int):
        classroom_id = classroom
    else:
        raise ValueError("classroom must be either int or Classroom")

    assignments: List[Assignment] = []

    assignments_d = _gh_api(f"classrooms/{classroom_id}/assignments")
    if not assignments_d:
        return []
    for assn_d in assignments_d:
        try:
            assignment = from_dict(Assignment, assn_d)
        except:
            logger.exception("Error converting Assignment")
            return None
        assignments.append(assignment)
    return assignments


@overload
def choose_assignment(classroom: int) -> Assignment: ...


@overload
def choose_assignment(classroom: Classroom) -> Assignment: ...


def choose_assignment(
    classroom: Classroom | int, suppress_print: bool = False
) -> Assignment:
    if isinstance(classroom, Classroom):
        classroom_id = classroom.id
    elif isinstance(classroom, int):
        classroom_id = classroom
    else:
        raise ValueError("classroom must be either int or Classroom")

    assignments = get_assignments(classroom_id)
    assignments.sort(key=lambda a: a.deadline if a.deadline else utc_min())
    assignments.reverse()
    maxlen = max([len(a.title) for a in assignments])
    choices = [
        a.title.ljust(maxlen + 2) + "Due: " + get_friendly_local_datetime(a.deadline)
        for a in assignments
    ]
    choice = select_choice(choices, title="CHOOSE ASSIGNMENT")
    if choice is not None:
        if not suppress_print:
            print(" > CHOICE:", choices[choice])
        return assignments[choice]
    else:
        if not suppress_print:
            print(" X CHOICE: NONE")
        return None


def get_assignment(assignment_id: int = None) -> Assignment:
    assn_d = _gh_api(f"assignments/{assignment_id}")
    if not assn_d:
        return None
    try:
        assignment = from_dict(Assignment, assn_d)
    except:
        logger.exception("Error converting Assignment")
        return None
    return assignment


###############################################################################
# get_accepted_assignments
###############################################################################


@overload
def get_accepted_assignments(assignment: int) -> List[AcceptedAssignment]: ...


@overload
def get_accepted_assignments(assignment: Assignment) -> List[AcceptedAssignment]: ...


def get_accepted_assignments(
    accepted_assignment: Assignment,
) -> List[AcceptedAssignment]:
    if isinstance(accepted_assignment, Assignment):
        assignment_id = accepted_assignment.id
    elif isinstance(accepted_assignment, int):
        assignment_id = accepted_assignment
    else:
        raise ValueError("assignment must be either int or Assignment")

    accepted_assignments: List[AcceptedAssignment] = []

    accepted_assignments_d = _gh_api(
        f"assignments/{assignment_id}/accepted_assignments"
    )
    if not accepted_assignments_d:
        return []
    for acc_assn_d in accepted_assignments_d:
        try:
            accepted_assignment = from_dict(AcceptedAssignment, acc_assn_d)
        except:
            logger.exception("Error converting AcceptedAssignment")
            return None
        accepted_assignments.append(accepted_assignment)
    return accepted_assignments


if __name__ == "__main__":
    c = choose_classroom()
    print(c)

    draw_single_line()

    a = choose_assignment(c)

    print(a)

    AcceptedAssignment.get_table_header()

    aa_list = get_accepted_assignments(a)
    for aa in aa_list:
        print(aa)



from pathlib import Path
from au import SettingsBase
from git_wrap import has_git_repo_subdirs, GitRepo


_CLASSROOM_ID = "Classroom.classroom_id"
_ASSIGNMENT_ID = "Assignment.assignment_id"
_ROSTER_FILE = "Classroom.roster_file"


class AssignmentSettings(SettingsBase):
    FILENAME = "assignment.toml"

    def __init__(self, settings_doc_path: Path | str, create=False):
        path = Path(settings_doc_path)
        if not path.exists():
            raise FileNotFoundError(f"{path} does not exist")
        file = path / self.FILENAME
        super().__init__(file, create)

    @property
    def classroom_id(self) -> int:
        return self.get(_CLASSROOM_ID)

    @classroom_id.setter
    def classroom_id(self, value):
        self.set(_CLASSROOM_ID, value)

    @property
    def assignment_id(self) -> int:
        return self.get(_ASSIGNMENT_ID)

    @assignment_id.setter
    def assignment_id(self, value):
        self.set(_ASSIGNMENT_ID, value)

    @property
    def roster_file(self) -> Path:
        return self.get(_ROSTER_FILE, is_path=True)

    @roster_file.setter
    def roster_file(self, value):
        self.set(_ROSTER_FILE, value)

    @staticmethod
    def is_valid_settings_path(path: Path) -> bool | None:
        """Returns True if this directory contains git repos. False if it IS a repo. None is indeterminate."""
        if GitRepo.is_repository_root(path):
            return False
        if has_git_repo_subdirs(path):
            return True
        return None

    @staticmethod
    def get_classroom_settings(
        settings_dir: Path | str, create: bool = False
    ) -> "AssignmentSettings":
        path = Path(settings_dir)
        try:
            return AssignmentSettings(path)
        except FileNotFoundError:
            try:
                return AssignmentSettings(path.parent)
            except FileNotFoundError:
                return AssignmentSettings(path, create=create)



import click
import pathlib

BASE_PATH_KEY = "au.base_path"


class BasePathType(click.Path):
    def __init__(self, exists=True, resolve_path=False, allow_dash=False):
        super().__init__(
            exists=exists,
            file_okay=False,
            dir_okay=True,
            path_type=pathlib.Path,
            resolve_path=resolve_path,
            allow_dash=allow_dash,
        )

    def convert(self, value, param, ctx):
        converted_value = super().convert(value, param, ctx)
        if isinstance(converted_value, pathlib.Path):
            ctx.meta[BASE_PATH_KEY] = converted_value.resolve()
            return converted_value
        return converted_value



#!/bin/env python3

import csv
import json
import os
import logging
import pprint
import difflib

from pathlib import Path

logging.basicConfig(level=logging.DEBUG)

source = Path(os.path.abspath(os.getcwd()))
logging.debug(source)

assignment_name = source.name
logging.debug(assignment_name)

csv_filename = source.joinpath('grades.csv')
logging.debug(csv_filename)

###############################################################################
# LOAD ALL RAW FEEDBACK
###############################################################################

all_feedback = []

# Assume '{assignment_name}-' prefixes each assignment to grade
for subdir in source.iterdir():
    if not subdir.is_dir():
        logging.debug(f'SKIPPING: {subdir} (not dir)')
        continue
    pos = len(assignment_name) + 1
    
    student_name = subdir.name.split('@')[0]

    feedback_file = subdir / '.sanity-check' / 'raw-feedback.json'

    if feedback_file.exists():
        try:
            with open(feedback_file, 'rt') as fi:
                feedback = json.load(fi)
        except:
            logging.error(f'UNABLE TO PROCESS {feedback_file}')

        if len(feedback) == 0:
            logging.info(f'SKIPPING {student_name} (no feedback)')
            continue

        feedback = [q for q in feedback if q['issues'] and q['issues'][0] and q['issues'][0]['issue'] != 'empty']
        if len(feedback) == 0:
            logging.info(f'SKIPPING {student_name} (no queries)')
            continue

        logging.debug(f"found feedback for {student_name}")
        all_feedback.append({
            'student_name': student_name,
            'feedback': feedback
        })

    else:
        logging.info(f"SKIPPING {student_name} (no raw-feedback.json)")

###############################################################################
# FIND SIMILARITIES
###############################################################################

checked = []

for student in all_feedback:
    for other in all_feedback:
        if student == other or (other,student) in checked:
            continue

        stu_name = student['student_name']
        oth_name = other['student_name']

        logging.debug(f"{stu_name} vs. {oth_name}")

        checked.append((student,other))

        stu_queries = {q['query_name']:q for q in student['feedback']}
        other_queries = {q['query_name']:q for q in other['feedback']}

        common_queries = [key for key in stu_queries if key in other_queries]
        different_queries = [key for key in stu_queries if key not in other_queries]
        different_queries += [key for key in other_queries if key not in stu_queries]
        common_count = len(common_queries)
        different_count = len(different_queries)

        common_query_details = {}

        logging.debug(f"common_count = {common_count}")

        # only continue if 'similar enough'
        if common_count == 0 or common_count <= different_count:
            continue
        
        identical_count = 0
        total_items = 0
        total_diff = 0

        identical_line_count = 0
        total_line_count = 0

        for q,stu_fb in stu_queries.items():
            if q in other_queries:
                other_fb = other_queries[q]

                stu_items = stu_fb['issues']
                other_items = other_fb['issues']

                shared_items = [item for item in stu_items if item in other_items]
                diff = abs(len(stu_items) - len(shared_items))

                if diff == 0:
                    identical_count += 1

                stu_sql = stu_fb['sql'].replace("'","").replace('"','').splitlines()
                stu_sql = [line.strip() for line in stu_sql]
                other_sql = other_fb['sql'].replace("'","").replace('"','').splitlines()
                other_sql = [line.strip() for line in other_sql]

                total_line_count += len(stu_sql)
                line_diff = difflib.ndiff(stu_sql, other_sql)
                diff_string = ''
                for d in line_diff:
                    diff_string += d + '\n'
                    if d.startswith('  '):
                        identical_line_count += 1


                common_query_details[q] = {
                    'stu_items': stu_items,
                    'other_items': other_items,
                    'stu_sql': stu_fb['sql'],
                    'other_sql': other_fb['sql'],
                    'diff': diff_string
                }

            else:
                diff = len(stu_items)

            total_diff += diff
            total_items += len(stu_items)

        common_issue_pct = round(100 * (1 - ((total_diff * 1.0) / total_items)))
        common_sql_pct = round(100 * ((identical_line_count * 1.0) / total_line_count))

        logging.debug(f"common_issue_pct = {common_issue_pct}, common_sql_pct = {common_sql_pct}")

        if common_issue_pct > 90 and common_sql_pct > 70:
            logging.info(f'SIMILAR: {student["student_name"]} <> {other["student_name"]}')
            logging.info(f'         common issues:    {common_issue_pct}% from {common_count} queries')
            logging.info(f'         common sql:       {common_sql_pct}%')
            # pprint.pprint(common_queries)

            issue_file = source.joinpath(student['roster_name'] + '--' + other['roster_name'] + '.similarity.txt')
            logging.debug(f'Writing {issue_file}')
            try:
                with open(issue_file, "w") as fi:
                    fi.write(f'STUDENT 1:       {student["student_name"]}\n')
                    fi.write(f'STUDENT 2:       {other["student_name"]}\n')
                    fi.write(f'COMMON ISSUES:   {common_issue_pct}% from {common_count} queries\n')
                    fi.write(f'COMMON SQL:      {common_sql_pct}%\n')

                    for q, data in common_query_details.items():
                        fi.write('-'*80 + '\n')
                        fi.write(q + '\n')
                        fi.write('-'*80 + '\n')
                        fi.write(data['diff'] + '\n')

            except Exception as ex:
                logging.error(f"Error writing {issue_file}: {ex}")




import click
import importlib
from pathlib import Path


class SubdirGroup(click.Group):
    def __init__(
        self, name=None, commands=None, file="", module="", cli_file="cli.py", **attrs
    ):
        self._file = file
        self._module = module
        self._cli_file = cli_file
        super().__init__(name, commands, **attrs)

    def list_commands(self, ctx: click.core.Context):
        rv = super().list_commands(ctx)
        module_dir = Path(self._file).parent
        for item in module_dir.iterdir():
            if item.is_dir() and item.name[0] not in "._":
                names = [f.name for f in item.iterdir()]
                if self._cli_file in names:
                    rv.append(item.name)
        rv.sort()
        return rv

    def get_command(self, ctx: click.core.Context, cmd_name: str):
        commands = self.list_commands(ctx)
        if cmd_name in commands:
            real_name = cmd_name
        else:
            matches = [x for x in commands if x.startswith(cmd_name)]
            if len(matches) != 1:
                return None
            real_name = matches[0]

        cmd = super().get_command(ctx, real_name)
        if cmd:
            return cmd
        else:
            try:
                mod = importlib.import_module(self._module + "." + real_name)
                return mod.cli.main
            except ImportError as ex:
                ctx.fail(f"Error importing tool module: {ex}")
                return None

    def resolve_command(self, ctx, args):
        # always return the full command name
        _, cmd, args = super().resolve_command(ctx, args)
        return cmd.name, cmd, args



from typing import List
from dataclasses import dataclass
from pathlib import Path
import re

_QUERY_FILE_PATTERN = re.compile(r"^([^\d]*)?(\d+)?(.*)?(.sql)$")


@dataclass
class QueryFile:
    file: Path
    prefix: str
    num: str
    suffix: str
    extension: str

    def query_name(self):
        return f"{self.prefix}{self.num}{self.suffix}"

    def file_name(self):
        return f"{self.prefix}{self.num}{self.suffix}{self.extension}"

    def sort_name(self, num_zeros: int) -> str:
        if not self.num:
            return self.query_name()
        diff = num_zeros - len(self.num)
        return self.prefix + "0" * diff + self.num + self.suffix

    @staticmethod
    def parse(file: Path):
        match = _QUERY_FILE_PATTERN.search(file.name)
        if not match:
            return None

        groups = match.groups()
        prefix = groups[0] if groups[0] else ""
        num = groups[1] if groups[1] else ""
        suffix = groups[2] if groups[2] else ""
        extension = groups[3] if groups[3] else ""
        return QueryFile(file, prefix, num, suffix, extension)


def get_query_files(path: Path) -> List[QueryFile]:
    queries: List[QueryFile] = []
    for file in path.iterdir():
        q = QueryFile.parse(file)
        if q:
            queries.append(q)

    if queries:
        max_q_len = max([len(q.num) for q in queries])
        queries.sort(key=lambda q: q.sort_name(max_q_len))

    return queries



from typing import Dict
import logging
import os
import sys
from io import StringIO
from pathlib import Path
from datetime import datetime, date
import json

import pytest
import pylint.lint as lint
from pylint.reporters.json_reporter import JSON2Reporter

import click
from rich.console import Console

from au.common import draw_double_line
from git_wrap import GitRepo, Commit

from au.classroom import Assignment, Roster
from au.click import BasePath, AssignmentOptions, RosterOptions, DebugOptions
from au.common import draw_single_line
from au.common.datetime import get_friendly_timedelta

from .pytest_reporter import PytestResultsReporter
from .scoring import get_summary


logger = logging.getLogger(__name__)


RESULTS_FILE_NAME = ".eval_results.json"


def _json_serialize(obj):
    if isinstance(obj, (datetime, date)):
        return obj.isoformat()


def _json_deserialize_hook(dct: Dict):
    for key, value in dct.items():
        if isinstance(value, str):
            try:
                dct[key] = datetime.fromisoformat(value.replace("Z", "+00:00"))
            except ValueError:
                pass
    return dct


def retrieve_student_results(student_dir: Path) -> Dict:
    results_file = student_dir / RESULTS_FILE_NAME
    with open(results_file, "r") as fi:
        return json.load(fi, object_hook=_json_deserialize_hook)


@click.command("eval-assignment")
@click.argument("student_dir", type=BasePath(), required=True)
@AssignmentOptions(store=False).options
@RosterOptions(store=False, prompt=True).options
@click.option(
    "--student-name",
    type=str,
    help="If no roster is provided, you can provide the name of the student. "
    "If neither is provided, the name will just be STUDENT_DIR.",
)
@DebugOptions().options
def eval_assignment_cmd(
    student_dir: Path,
    assignment: Assignment,
    roster: Roster = None,
    student_name: str = None,
    **kwargs,
) -> None:
    """Run automated grading tests on a single student directory."""
    logging.basicConfig()

    draw_double_line()
    if assignment:
        print(assignment)
    else:
        logger.error(
            "Unable to find the requested assignment. Functionality will be limited."
        )
    draw_single_line()

    if roster:
        print(f"USING ROSTER: {roster.file.resolve()}")
    else:
        print("NOT USING ROSTER")

    draw_single_line()
    print()

    stu_name = student_name
    if roster and not stu_name:
        login = roster.get_login_for_dir(student_dir)
        stu_name = roster.login_student_map.get(login)

    student_results = eval_assignment(student_dir, stu_name, assignment)

    if student_results:
        draw_single_line(f"Summary Results")
        print(get_summary(student_results))
    else:
        print(f"No results for {stu_name}")


def eval_assignment(
    student_dir: Path, student_name: str, assignment: Assignment = None
) -> Dict[str, any]:
    """
    Run automated grading tests on a single student directory.
    """
    dir_name = student_dir.name
    os.chdir(student_dir)
    student_results: dict = {}

    student_results["name"] = student_name
    student_results["dir_name"] = dir_name
    if assignment:
        student_results["assignment_title"] = assignment.title
        student_results["assignment_deadline"] = assignment.deadline

    ###############################################################################
    # REPO CHECKS
    ###############################################################################

    console = Console()
    with console.status(
        f"Running Git Repository checks for {student_name}", spinner="bouncingBar"
    ) as status:
        try:
            # repo = Repo()
            repo = GitRepo()
            assert repo
        except:
            logger.error(f"No git repository found in {dir_name}")
            return None

        self_email = GitRepo.get_user_email()
        commits: list[Commit] = []
        for commit in repo.get_commits():
            # Skip the evaluator...best effort
            if commit.author_email == self_email:
                continue
            # First GitHub commit is considered the end of the list of student
            # commits
            if commit.committer_name == "GitHub":
                break
            commits.append(commit)

        student_results["num_commits"] = len(commits)
        if commits:
            commit = commits[0]
            if not student_name:
                student_results["name"] = commit.author_name
            student_results["commit_message"] = commit.message.strip()
            student_results["commiter_name"] = commit.author_name
            commit_date = commit.date
            student_results["commit_date"] = commit_date
            if assignment:
                past_due = commit_date - assignment.deadline
                if past_due.total_seconds() > 0:
                    student_results["past_due"] = get_friendly_timedelta(past_due)
        else:
            return None

    ###############################################################################
    # PYTEST
    ###############################################################################

    draw_single_line("pytest")

    pytest_reporter = PytestResultsReporter()

    keep_packages = [
        "_asyncio",
        "_contextvars",
        "_elementtree",
        "_pytest",
        "_ssl",
        "asyncio",
        "attr",
        "cmd",
        "code",
        "codeop",
        "contextvars",
        "faulthandler",
        "pdb",
        "pkgutil",
        "pyexpat",
        "pytest_metadata",
        "pytest_subtests",
        "readline",
        "ssl",
        "unittest",
        "xml",
    ]

    pretest_modules = [key for key in sys.modules.keys()]
    pretest_path = sys.path.copy()

    try:
        # run the tests and report
        pytest.main(
            [
                "--tb=no",
                "-q",
            ],
            plugins=[pytest_reporter],
        )

        pytest_pct = round(pytest_reporter.results.pass_pct, 3)
        student_results["pytest_pct"] = pytest_pct
        student_results["pytest_results"] = pytest_reporter.results.as_dict()

    except Exception as ex:
        student_results["pytest_exception"] = ex
        logger.exception("Unexpected error running pytest")

    finally:
        posttest_modules = [key for key in sys.modules.keys()]
        for module_name in posttest_modules:
            if module_name in pretest_modules:
                continue
            package = module_name.split(".")[0]
            if package in keep_packages:
                continue
            del sys.modules[module_name]
        sys.path = pretest_path.copy()

    ###############################################################################
    # PYLINT
    ###############################################################################

    draw_single_line("pylint")
    status.update(status=f"Running pylint for {student_name}")

    lint_files = []
    lint_filenames = []
    cwd = Path.cwd()
    for root, dirs, files in os.walk(cwd):
        dirs[:] = [d for d in dirs if d[0] not in "._" and d[:4] != "test"]
        for file in files:
            if (
                file.endswith(".py")
                and not file.startswith(".")
                and not file.startswith("_")
                and not file.startswith("test_")
                and not file.endswith("_test.py")
            ):
                lint_filenames.append(file)
                lint_files.append(root + os.sep + file)

    if lint_files:
        print("Testing", *lint_filenames)

        lint_disabled = [
            "invalid-name",
            "missing-module-docstring",
            "missing-class-docstring",
            "missing-function-docstring",
            "trailing-whitespace",
            "missing-final-newline",
            "trailing-newlines",
            "unnecessary-negation",
            "wrong-import-order",
            "duplicate-code",
            "too-few-public-methods",
            "too-many-arguments",
            "too-many-locals",
            "too-many-statements",
            "bare-except",
            "f-string-without-interpolation",
            "chained-comparison",
            "consider-using-sys-exit",
            "singleton-comparison",
            "consider-using-max-builtin",
            # 'bad-indentation',
            "redefined-outer-name",
            "simplifiable-if-statement",
            "no-else-return",
            "inconsistent-return-statements",
        ]
        lint_args = []
        lint_args += ["--disable=" + ",".join(lint_disabled)]
        lint_args += lint_files
        pylint_output = StringIO()  # Custom open stream
        pylint_reporter = JSON2Reporter(pylint_output)

        try:
            lint.Run(lint_args, reporter=pylint_reporter, exit=False)
            pylint_results = json.loads(pylint_output.getvalue())
            pylint_pct = round(pylint_results["statistics"]["score"] / 10.0, 3)
            student_results["pylint_pct"] = pylint_pct
            if pylint_pct < 1:
                student_results["pylint_results"] = pylint_results
        except Exception as ex:
            student_results["pylint_exception"] = ex
            logger.exception("Unexpected error running pylint")
    else:
        logger.error("No files found to lint found")

    ###############################################################################
    # Save and return the test results data
    ###############################################################################

    with open(RESULTS_FILE_NAME, "w") as fi:
        json.dump(student_results, fi, indent=2, default=_json_serialize)

    return student_results


if __name__ == "__main__":
    eval_assignment_cmd()



#!/usr/bin/python3

from commonsql import parseSQL, getProperties, resetDatabase
from commonsql import logger, baseLoggingLevel
from commonutil import fullpath
from pprint import pformat as pf
import logging
import props

import json
import os
import re


_debugQueries = []

_databases = {}
_queries_full = {}
_queries_sanity = {}

while True:
    dbnum = len(_databases) + 1
    dbname = input(f'Enter name of database {dbnum} (enter to stop): ').strip().lower()
    if len(dbname) == 0:
        break
    dbfilename = input(f'Enter schema filename ({dbname}.sql): ').strip()
    if len(dbfilename) == 0:
        dbfilename = dbname + '.sql'
    _databases[dbname] = dbfilename

if len(_databases) == 0:
    exit(0)

for dbname, dbfilename in _databases.items():
    file = fullpath(dbfilename)
    resetDatabase(dbname, file)

# This mess of convolution is needed solely to sort the files
# in a logical order.

queryfiles = []
maxnums = 0
for filename in os.listdir():
    if not filename.endswith('.sql'):
        continue

    query_name = filename[:-4]
    match = re.search('(\d+)(.*)',query_name)
    if not match:
        continue

    numcount = len(match.group(1))
    if numcount > maxnums:
        maxnums = numcount

    sortname = match.group(0)
    
    queryfiles.append([sortname, query_name, numcount])

for i in range(len(queryfiles)):
    numcount = queryfiles[i][2]
    numdiff = maxnums - numcount
    if numdiff:
        queryfiles[i][0] = ('0' * numdiff) + queryfiles[i][0]

# Now done with allowing for sorting, cycle through the sorted list of queries

for _, query_name, _ in sorted(queryfiles):

    debug_query = query_name in _debugQueries
    if debug_query:
        logger.setLevel(logging.DEBUG)
        logger.debug(f"Forcing DEBUG level on {query_name}")
    else:
        logger.setLevel(baseLoggingLevel)

    logger.info(f'Processing {query_name}')
    properties_full = {}

    try:
        with open(query_name + '.sql') as fi:
            sql = parseSQL(fi.read(), properties_full)

        if len(sql) == 0:
            logger.warning(f"No SQL found for {query_name}")
            continue

        if len(_databases) == 1:
            dbname = list(_databases.keys())[0]
        else:
            if props.DATABASE_NAME in properties_full:
                dbname = properties_full[props.DATABASE_NAME]
            else:
                dbname = input("Enter dbname: ").lower().strip()

        properties_full[props.DATABASE_NAME] = dbname
        properties_full[props.SOLUTION] = sql
        getProperties(sql, properties_full)

    except OSError as ex:
        logger.exception(f"Error reading SQL file")
        exit(1)
    except Exception as ex:
        logger.exception(f"Error executing SQL")
        exit(1)

    _queries_full[query_name] = properties_full
    properties_sanity = {}
    for propname in props.SANITY_PROPS:
        if propname in properties_full.keys():
            properties_sanity[propname] = properties_full[propname]
    _queries_sanity[query_name] = properties_sanity

jsondata_full = {
    props.ALL_DATABASES : _databases,
    props.ALL_QUERIES : _queries_full
}
jsonText_full = json.dumps(jsondata_full, indent=2)

logger.debug("Full Check:\n" + jsonText_full)

jsondata_sanity = {
    props.ALL_DATABASES : _databases,
    props.ALL_QUERIES : _queries_sanity
}
jsonText_sanity = json.dumps(jsondata_sanity, indent=2)

logger.debug("Sanity Check:\n" + jsonText_sanity)

path = os.path.dirname(os.path.realpath(__file__))

jsonFilename = path + '/' + props.FULL_CHECK_FILENAME
with open(jsonFilename, mode='w') as fi:
    fi.write(jsonText_full)
print(f'Wrote {len(_queries_full)} queries to {jsonFilename}')

jsonFilename = path + '/' + props.SANITY_CHECK_FILENAME
with open(jsonFilename, mode='w') as fi:
    fi.write(jsonText_sanity)

logger.info(f'Wrote {len(_queries_full)} queries to {jsonFilename}')



#!/usr/bin/python3

###############################################################################
# 
#   D O   N O T   M O D I F Y   T H I S   F I L E
# 
###############################################################################
# 
# You can run it from the terminal by entering:
# 
#      ./checkit
# 
# and pressing enter. Or just hit the 󠁆▷ button to run it if you're simply not
# cool enough to type terminal commands.
# 
###############################################################################

from commonutil import ps, pb, fullpath
from commonsql import resetDatabase, parseSQL, getProperties
import props
import json

ps([
    'SANITY CHECK',
    '============',
    '''
        This script tests only basic properties of your SQL queries. It is meant
        only to prevent major blunders. Any given test may "pass" and you may
        still have significant errors in your SQL.
    '''
])
print()

with open(fullpath(props.SANITY_CHECK_FILENAME)) as fi:
    querydata = json.load(fi)

databases = querydata[props.ALL_DATABASES]

for dbname, dbfilename in databases.items():
    print(f"Resetting {dbname} database...")
    file = fullpath(dbfilename)
    resetDatabase(dbname, file)

print(f"Executing tests...")

queries = querydata[props.ALL_QUERIES]
queryissues = {}

for queryname, expectedprops in queries.items():

    issues = []
    actualprops = dict(expectedprops)

    try:
        with open(queryname + '.sql') as fi:
            sql = parseSQL(fi.read(), actualprops)

        if len(sql) == 0:
            verb = actualprops.setdefault(props.SQL_VERB, None)
            if verb:
                issues.append(f"Unknown SQL verb {verb.upper()}")
            else:
                issues.append("EMPTY")

        else:

            # copy expectedprops to fill in
            getProperties(sql, actualprops)
        
            # print(expectedprops)
            verb = expectedprops['verb']

            for propname, exp in expectedprops.items():
                exp = expectedprops[propname]
                act = actualprops[propname]
                friendly = props.friendly(propname)
                if exp != act:
                    issues.append(f"Expected {exp} for {friendly} but retrieved {act}")

    except OSError as ex:
        issues.append(f"Error reading SQL file: {ex}")
    except Exception as ex:
        issues.append(f"Error executing SQL: {ex}")

    queryissues[queryname] = issues


print()

total = len(queries)
passed = 0
failed = 0
empty = 0

for queryname, issues in queryissues.items():
    if len(issues) == 0:
        print(f"  ✓  {queryname}")
        passed += 1
    else:
        if issues[0] == "EMPTY":
            print(f"  _  {queryname} (NO SQL PROVIDED)")
            empty += 1
        else:
            print(f"  ✗  {queryname}")
            failed += 1
            pb(issues)

print()
print( '╔══════════════════════╗')
print(f"║ queries checked: {total:3} ║")
print( '╟──────────────────────╢')
print(f"║          passed: {passed:3} ║")
print(f"║          failed: {failed:3} ║")
print(f"║           empty: {empty:3} ║")
print( '╚══════════════════════╝')
print()







from .classroom_api import choose_classroom
from .classroom_api import get_classrooms
from .classroom_api import get_classroom
from .classroom_api import choose_assignment
from .classroom_api import get_assignments
from .classroom_api import get_assignment
from .classroom_api import get_accepted_assignments

from .classroom_types import Organization
from .classroom_types import Classroom
from .classroom_types import Assignment
from .classroom_types import AcceptedAssignment
from .classroom_types import Repository
from .classroom_types import Student

from .AssignmentSettings import AssignmentSettings

from .Roster import Roster



from typing import Optional, List
from dataclasses import dataclass, asdict
from datetime import date, datetime
import json
from pprint import pformat

from au.classroom import Roster
from f_table import get_table

from au.common.datetime import date_to_local, get_friendly_local_datetime


def _github_json_serializer(obj):
    if isinstance(obj, (date, datetime)):
        dt = date_to_local(obj)
        return dt.isoformat()
    raise TypeError(f"Type {type(obj)} not JSON serializable")


def _github_json_deserializer(dct):
    for key, value in dct.items():
        if isinstance(value, str):
            try:
                dct[key] = datetime.fromisoformat(value.replace("Z", "+00:00"))
            except ValueError:
                pass
    return dct


def _as_json(obj):
    dct = asdict(obj)
    return json.dumps(dct, default=_github_json_serializer)


@dataclass
class Organization:
    id: int
    login: str
    node_id: Optional[str]
    html_url: Optional[str]
    name: Optional[str]
    avatar_url: Optional[str]

    def as_table(self) -> str:
        return get_table(
            [
                ["Organization Login", self.login],
                ["Organization ID", self.id],
                ["Organization Name", self.name],
                ["Organization Node ID", self.node_id],
                ["Organization HTML URL", self.html_url],
                ["Organization Avatar URL", self.avatar_url],
            ],
            col_defs=["25", ""],
        )

    def __str__(self):
        return self.as_table()


@dataclass
class Classroom:
    id: int
    name: str
    url: str
    archived: bool
    organization: Optional[Organization]

    def as_table(self) -> str:
        table = get_table(
            [
                ["Classroom Name", self.name],
                ["Classroom ID", self.id],
                ["Classroom URL", self.url],
                ["Classroom is Archived", self.archived],
            ],
            col_defs=["25", ""],
        )
        if self.organization:
            table += "\n" + self.organization.as_table()
        return table

    def as_json(self) -> str:
        return _as_json(self)

    def __str__(self):
        return get_table(
            [
                ["Classroom Name", self.name],
                ["Classroom ID", self.id],
            ],
            col_defs=["25", ""],
        )


@dataclass
class Repository:
    id: int
    name: str
    full_name: str
    html_url: str
    node_id: str
    private: bool
    default_branch: str

    def as_json(self) -> str:
        return _as_json(self)

    def __str__(self):
        return pformat(self)


@dataclass
class Assignment:
    id: int
    title: str
    slug: Optional[str]
    deadline: Optional[datetime]
    accepted: Optional[int]
    submissions: Optional[int]
    passing: Optional[int]
    invite_link: Optional[str]
    type: Optional[str]
    editor: Optional[str]
    public_repo: Optional[bool]
    invitations_enabled: Optional[bool]
    students_are_repo_admins: Optional[bool]
    feedback_pull_requests_enabled: Optional[bool]
    max_teams: Optional[int]
    max_members: Optional[int]
    language: Optional[str]
    classroom: Optional[Classroom]
    starter_code_repository: Optional[Repository]

    def as_table(self) -> str:
        table = get_table(
            [
                ["Classroom Name", self.classroom.name],
                ["Classroom ID", self.classroom.id],
                ["Assignment Title", self.title],
                ["Short Name (slug)", self.slug],
                ["Assignment ID", self.id],
                ["Assignment Deadline", get_friendly_local_datetime(self.deadline)],
                ["Assignment Type", self.type],
                ["Assignment Editor", self.editor],
                ["Assignment is Public", self.public_repo],
                ["Number Accepted", self.accepted],
                ["Number Submitted", self.submissions],
                ["Number Passing", self.passing],
            ],
            col_defs=["25", ""],
        )
        return table

    def as_json(self) -> str:
        return _as_json(self)

    def __str__(self):
        table = get_table(
            [
                ["Assignment Title", self.title],
                ["Assignment ID", self.id],
                ["Deadline", get_friendly_local_datetime(self.deadline)],
            ],
            col_defs=["25", ""],
        )
        if self.classroom:
            table = str(self.classroom) + "\n" + table
        return table


@dataclass
class Student:
    id: int
    login: str
    name: Optional[str]
    avatar_url: str
    html_url: str

    def as_json(self) -> str:
        return _as_json(self)

    def __str__(self):
        return pformat(self)


@dataclass
class AcceptedAssignment:
    id: int
    submitted: bool
    passing: bool
    commit_count: int
    grade: Optional[str]
    students: List[Student]
    assignment: Assignment
    repository: Repository

    @property
    def login(self):
        if not self.students:
            return None
        else:
            return self.students[0].login

    def as_json(self) -> str:
        return _as_json(self)

    def __str__(self):
        return pformat(self)



from .git_repo import GitRepo
from .git_repo import Commit

from .git_repo import has_git_repo_subdirs
from .git_repo import get_git_dirs
from .git_repo import get_git_repos
from .git_repo import get_dirty_repos




import click
from au.click import AliasedGroup
from importlib.metadata import version, PackageNotFoundError

from .assignment.cli import assignment
from .python.cli import python
from .sql.cli import sql


MODULE_NAME = "au-tools"
try:
    VERSION = version(MODULE_NAME)
except PackageNotFoundError:
    VERSION = "undetermined"


# @click.command(cls=SubdirGroup, file=__file__, module=__package__)
@click.version_option(VERSION)
@click.group(cls=AliasedGroup)
def main():
    """
    au - GitHub Classroom Automation Tools

    The gold standard for managing GitHub Classroom assignments at scale. These
    tools help to automate the workflows required to create, administer,
    evaluate, and provide feedback for assignments.

    Homepage: https://ptyork.github.io/au
    """


main.add_command(assignment)
main.add_command(python)
main.add_command(sql)


if __name__ == "__main__":
    main()



from os import get_terminal_size

MAX_REASONABLE_WIDTH = 120


def get_term_width(max_term_width: int = MAX_REASONABLE_WIDTH):
    try:
        term_width = get_terminal_size().columns
    except:
        return max_term_width
    if max_term_width == 0:
        max_term_width = 9999
    return min(max_term_width, term_width)



#!/usr/bin/python3

import csv
import os
import logging
import pprint

from pathlib import Path

logging.basicConfig(level=logging.DEBUG)

source = Path(os.path.abspath(os.getcwd()))
logging.debug(source)

assignment_name = source.name
logging.debug(assignment_name)

csv_filename = source / 'grades.csv'
logging.debug(csv_filename)

scores = []
score_prefix = '## SCORE:'

###############################################################################
# FIND ALL ASSIGNMENTS
###############################################################################

assignments = []

# Assume '{assignment_name}-' prefixes each assignment to grade
for subdir in source.iterdir():
    if not subdir.is_dir():
        logging.debug(f'SKIPPING: {subdir} (not dir)')
        continue
    feedback_file = subdir / 'FEEDBACK.md'
    if not feedback_file.exists():
        logging.info(f'SKIPPING: {subdir} (no feedback)')
        continue
    
    student_name = subdir.name.split('@')[0]

    logging.debug(f"found feedback for {student_name}")

    ###############################################################################
    # EXTRACT ALL SCORES
    ###############################################################################

    score = ''

    try:
        with open(feedback_file, 'rt') as fi:
            for line in fi:
                if line.startswith(score_prefix):
                    score = line.replace(score_prefix, '').strip()
                    logging.info(f'SCORE for {student_name}: {score}')
                    break
    except:
        logging.error(f'UNABLE TO PROCESS {feedback_file}')
    if not score:
        logging.info(f'SKIPPING {student_name} (no score)')

    scores.append([student_name, score])

scores.sort()

logging.debug(pprint.pformat(scores, indent=4))

###############################################################################
# WRITE CSV
###############################################################################

try:
    with open(csv_filename, "w") as csv_file:
        writer = csv.writer(csv_file)
        writer.writerow(('student','score'))
        writer.writerows(scores)
except Exception as ex:
    logging.error(f"Error reading {csv_filename}: {ex}")



import click
import sys
import functools
import logging
from pathlib import Path

from au.classroom import AssignmentSettings, Roster
from au.common import select_file
from .BasePathType import BASE_PATH_KEY

logger = logging.getLogger(__name__)


class RosterOptions:
    def __init__(
        self,
        load: bool = True,
        store: bool = True,
        prompt=False,
        required=False,
    ):
        self.load = load
        self.store = store
        self.prompt = prompt
        self.required = required

    def _get_settings(self, base_path):
        if base_path and (self.load or self.store):
            try:
                return AssignmentSettings.get_classroom_settings(base_path)
            except:
                if self.store:
                    if AssignmentSettings.is_valid_settings_path(base_path):
                        return AssignmentSettings(base_path, create=True)
        elif self.load or self.store:
            logger.error(
                "RosterOptions can't be used with store or load if an argument of type BasePath is not also included."
            )
            sys.exit(1)
        return None

    def options(self, func):
        help_text = (
            "A GitHub Classroom roster file, typically named `classroom_roster.csv`."
        )
        if self.load:
            help_text += (
                " If not provided, will try to read a value stored in settings."
            )
        if self.prompt:
            help_text += " Will prompt for the file if not provided."
        if self.required:
            help_text += " A roster is REQUIRED to run this command."

        @click.option(
            "--roster",
            type=click.Path(dir_okay=False, exists=True, path_type=Path),
            help=help_text,
        )
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            ctx = click.get_current_context()
            base_path = ctx.meta.get(BASE_PATH_KEY)  # Command must use BasePath type
            settings = self._get_settings(base_path)
            roster_file: Path = None
            if "roster" in kwargs:
                roster_file = kwargs.get("roster")
            kwargs["roster"] = None  # replace with the actual roster if found
            if self.load and not roster_file:
                if settings and settings.roster_file:
                    roster_file = settings.roster_file
            if not roster_file and self.prompt:
                optional = ""
                if not self.required:
                    optional = " (Press esc for none)"
                roster_file = select_file(
                    title=f"CHOOSE ROSTER CSV FILE{optional}",
                    root=base_path,
                    filter="*.csv",
                )
                if roster_file:
                    click.echo(f" > CHOICE: {roster_file}")
                else:
                    click.echo(" X CHOICE: NONE")
            if not roster_file and self.required:
                click.echo("A roster is required for this operation.\n")
                click.echo(click.get_current_context().get_help())
                sys.exit(1)
            if roster_file:
                try:
                    roster = Roster(roster_file)
                    kwargs["roster"] = roster
                except:
                    logger.exception("Error encountered while processing roster")
                    sys.exit(1)
                try:
                    if settings and self.store:
                        with settings:
                            settings.roster_file = roster_file
                except Exception:
                    logger.exception("Error encountered while writing settings")
            return func(*args, **kwargs)

        return wrapper



from pathlib import Path
from pydoc import pager

import click

from au.click import AliasedGroup, BasePath
from au.common import get_double_line

from .query_file import get_query_files


@click.command(cls=AliasedGroup)
def sql():
    """Commands for working with SQL assignments."""
    pass


@sql.command()
@click.argument("path", type=BasePath())
@click.option(
    "-pc",
    "--preserve-comments",
    is_flag=True,
    help="Disable removal of single-line comments.",
)
def cat_sql(path: Path, preserve_comments: bool):
    """
    Concatenate all text from *.sql files and show them in a pager.
    """
    queries = get_query_files(path)
    if queries:
        all_sql = ""
        for query in queries:
            all_sql += get_double_line(query.query_name()) + "\n"
            with open(query.file, "rt") as fi:
                for line in fi:
                    if preserve_comments or (
                        line.strip() and not line.lstrip().startswith("--")
                    ):
                        all_sql += line
            all_sql += "\n"
        pager(all_sql)
    else:
        click.echo(f"\nNo SQL files found in {path}\n")
        # click.echo(click.get_current_context().get_help())


if __name__ == "__main__":
    sql()



from typing import Iterator, List, Iterable, SupportsIndex, overload
from dataclasses import dataclass
import re
from textwrap import wrap

from .table_styles import TableStyle, NoBorderScreenStyle
from .terminal import get_term_width


class InvalidTableError(ValueError): ...


class InvalidColDefError(ValueError): ...


###############################################################################
# ColDef
###############################################################################

_FORMAT_SPEC_PATTERN = re.compile(
    r"((?P<fill>.)?(?P<align>[<>=^]))?"
    r"(?P<sign>[+\- ])?"
    r"(?P<alternate>[z#])?"
    r"(?P<zero>0)?"
    r"(?P<width>\d+)?"
    r"(?P<grouping_option>[_,])?"
    r"(?P<precision>\.\d+)?"
    r"(?P<type>[bcdeEfFgGnosxX%])?"
    r"(?P<table_config>[ANTS]+)?"
)


@dataclass
class ColDef:
    width: int | None = None
    align: str = "<"
    auto_fill: bool = False
    wrap: bool = True
    truncate: bool = False
    strict: bool = False
    _format_spec: str = ""

    def get_format_string(self) -> str:
        if self._format_spec:
            return f"{{:{self._format_spec}}}"
        else:
            return self.get_fallback_format_string()

    def get_fallback_format_string(self) -> str:
        return f"{{:{self.align}{self.width}}}"

    def format(self, value: any) -> str:
        # "Inner" format
        try:
            format_string = f"{{:{self._format_spec}}}"
            text = format_string.format(value)
        except:
            if self.strict:
                raise
            else:
                text = str(value)

        # "Outer" format
        return self.format_text(text)

    def format_text(self, text: str) -> str:
        if len(text) > self.width and self.truncate:
            text = text[: self.width - 1] + "…"

        if self.align == "^":
            text = text.strip().center(self.width)
        elif self.align == ">":
            text = text.strip().rjust(self.width)
        else:
            text = text.ljust(self.width)

        return text

    @staticmethod
    def parse(text) -> "ColDef":
        match = _FORMAT_SPEC_PATTERN.match(text)
        if not match:
            raise InvalidColDefError(f"Invalid format specifier for column: {text}")
        spec = match.groupdict()
        align = spec["align"]
        if not align or align == "=":
            align = ""
        width = spec["width"]
        if not width:
            width = None
        else:
            width = int(width)

        auto_size = False
        wrap_line = True
        truncate = False

        table_config = spec["table_config"]
        if table_config:
            if "A" in table_config:
                auto_size = True
            if "N" in table_config:
                wrap_line = False
            if "T" in table_config:
                truncate = True
            format_spec = text.removesuffix(table_config)
        else:
            format_spec = text

        # if format spec is just a number, then just toss it to avoid
        # inadvertent right-aligned numbers.
        try:
            _ = int(format_spec)
            format_spec = ""
        except ValueError:
            pass

        return ColDef(
            width=width,
            align=align,
            auto_fill=auto_size,
            wrap=wrap_line,
            truncate=truncate,
            _format_spec=format_spec,
        )


###############################################################################
# ColDefList
###############################################################################


class ColDefList(list[ColDef]):
    """
    A list of ColDef objects.
    """

    def __init__(self, iterable: Iterable = []):
        super().__init__()
        for val in iterable:
            self.append(val)

    @overload
    def __setitem__(self, key: SupportsIndex, value: str | ColDef, /) -> None: ...

    @overload
    def __setitem__(self, key: slice, value: Iterable[str | ColDef], /) -> None: ...

    def __setitem__(self, key, value):
        if isinstance(key, SupportsIndex):
            if isinstance(value, str):
                super().__setitem__(key, ColDef.parse(value))
            elif isinstance(value, ColDef):
                super().__setitem__(key, value)
            else:
                raise ValueError("Column definition contain an invalid value")
        elif isinstance(key, slice) and isinstance(value, Iterable):
            values = ColDefList(value)
            super().__setitem__(key, values)
        else:
            raise ValueError("Column definitions contain an invalid value")

    @overload
    def __getitem__(self, i: SupportsIndex) -> ColDef: ...

    @overload
    def __getitem__(self, i: slice) -> "ColDefList": ...

    def __getitem__(self, i):
        result = super().__getitem__(i)
        if isinstance(i, slice):
            return ColDefList(result)
        else:
            return result

    def __iter__(self) -> Iterator[ColDef]:
        return super().__iter__()

    def append(self, object):
        if isinstance(object, str):
            super().append(ColDef.parse(object))
        elif isinstance(object, ColDef):
            super().append(object)
        else:
            raise ValueError("Column definitions contain an invalid value")

    def adjust_to_table(
        self,
        table_data: List[List[any]],
        table_width: int,
        style: TableStyle,
    ) -> None:
        # ADD MISSING COL DEFS
        max_cols = max([len(row) for row in table_data])
        diff = max_cols - len(self)
        if diff:
            for _ in range(diff):
                self.append(ColDef())

        # ADJUST WIDTHS OF FIELDS TO MATCH REALITY
        for col_idx in range(max_cols):
            max_width = max([len(str(row[col_idx])) for row in table_data])
            col_def = self[col_idx]
            if not col_def.width:
                col_def.width = max_width
            if col_def.width < style.min_width:
                col_def.width = style.min_width

        # ADJUST AUTO-FILL COLS TO FILL REMAINING SPACE AVAILABLE IN TOTAL TABLE_WIDTH
        if not table_width:
            return

        padding_len = style.cell_padding * 2 * len(self)
        border_len = len(str(style.values_left)) + len(str(style.values_right))
        delims_len = len(str(style.values_delimiter)) * (len(self) - 1)
        non_text_len = padding_len + border_len + delims_len
        total_len = non_text_len + sum([c.width for c in self])

        fill_cols = [col_idx for col_idx in range(len(self)) if self[col_idx].auto_fill]
        if not fill_cols:
            if total_len <= table_width:
                return
            else:
                largest_col = self[0]
                largest_col_idx = 0
                for col_idx in range(1, len(self)):
                    col_def = self[col_idx]
                    if col_def.width > largest_col.width:
                        largest_col = col_def
                        largest_col_idx = col_idx
                largest_col.auto_fill = True
                fill_cols.append(largest_col_idx)

        fixed_len = sum([c.width for c in self if not c.auto_fill])

        remaining_width = table_width - fixed_len - non_text_len
        fill_width = remaining_width // len(fill_cols)

        if fill_width < style.min_width:
            raise ValueError(
                "Unable to expand columns to fit table width because existing columns are too wide"
            )

        remainder = remaining_width % len(fill_cols)
        for col_idx in fill_cols:
            new_width = fill_width
            if remainder:
                new_width += 1
                remainder -= 1
            self[col_idx].width = new_width

    @staticmethod
    def assert_valid_table(table: any) -> int:
        if not isinstance(table, (list, tuple)):
            raise ValueError("Table data must be a list or tuple")
        for row in table:
            if not isinstance(row, (list, tuple)):
                raise ValueError("Each row in a table must be a list or tuple")
            for cell in row:
                if isinstance(cell, (list, tuple, dict)):
                    raise ValueError(
                        "Each cell in a table row must contain a single value (not a list, tuple or dict)."
                    )

    @staticmethod
    def for_table(table: List[List[any]]) -> "ColDefList":
        ColDefList.assert_valid_table(table)
        max_cols = max([len(row) for row in table])
        col_defs = ColDefList([ColDef() for _ in range(max_cols)])
        for col_idx in range(max_cols):
            max_width = 0
            for row in table:
                if col_idx < len(row):
                    max_width = max(max_width, len(str(row[col_idx])))
            col_defs[col_idx].width = max_width
        return col_defs


###############################################################################
# get_table_row
###############################################################################


def _get_table_row(
    values: List[str],
    style: TableStyle = NoBorderScreenStyle(),
    col_defs: List[str | ColDef] | ColDefList = None,
    table_width: int = None,
    lazy_end: bool = False,
    is_header: bool = False,
) -> str:
    if not table_width and style.terminal_style:
        table_width = get_term_width()

    if not col_defs:
        _col_defs = ColDefList.for_table([values])
    elif isinstance(col_defs, ColDefList):
        _col_defs = col_defs
    else:
        _col_defs = ColDefList(col_defs)
    _col_defs.adjust_to_table([values], table_width, style)

    col_count = len(values)

    formatted_values = []
    for col_idx in range(col_count):
        col_val = values[col_idx]
        col_def = _col_defs[col_idx]
        text = col_def.format(col_val)
        formatted_values.append(text)

    all_col_lines = []
    for col_idx in range(col_count):
        col_lines = []
        col_def = _col_defs[col_idx]
        text = formatted_values[col_idx]
        split = text.splitlines()
        if not split:
            split = [""]
        for line in split:
            wrapped = wrap(line, width=col_def.width)
            if wrapped:
                for wrapped_line in wrapped:
                    col_lines.append(wrapped_line)
            else:
                col_lines.append(line)
            all_col_lines.append(col_lines)

    max_rows = max([len(col) for col in all_col_lines])

    if max_rows == 1:
        wrapped_rows = [formatted_values]
    else:
        wrapped_rows = []
        for row_idx in range(max_rows):
            row = []
            for col_idx in range(col_count):
                col = all_col_lines[col_idx]
                col_def = _col_defs[col_idx]
                if row_idx < len(col):
                    text = col[row_idx]
                else:
                    text = ""
                text = col_def.format_text(text)
                row.append(text)
            wrapped_rows.append(row)

    padding = " " * style.cell_padding
    if is_header:
        delim = padding + str(style.header_delimiter) + padding
        left = str(style.header_left) + padding
        right = "" if lazy_end else padding + str(style.header_right)
    else:
        delim = padding + str(style.values_delimiter) + padding
        left = str(style.values_left) + padding
        right = "" if lazy_end else padding + str(style.values_right)

    final_rows = []
    for row in wrapped_rows:
        row_text = left + delim.join(row) + right
        if lazy_end:
            row_text = row_text.rstrip()
        final_rows.append(row_text)

    return "\n".join(final_rows)


def get_table_row(
    values: List[str],
    style: TableStyle = NoBorderScreenStyle(),
    col_defs: List[str | ColDef] | ColDefList = None,
    table_width: int = None,
    lazy_end: bool = True,
) -> str:
    return _get_table_row(
        values=values,
        style=style,
        col_defs=col_defs,
        table_width=table_width,
        lazy_end=lazy_end,
        is_header=False,
    )


###############################################################################
# get_table_header
###############################################################################


def get_table_header(
    header_cols: List[str],
    style: TableStyle = NoBorderScreenStyle(),
    header_defs: List[str | ColDef] | ColDefList = None,
    col_defs: List[str | ColDef] | ColDefList = None,
    table_width: int = None,
    lazy_end: bool = True,
) -> str:
    if not table_width and style.terminal_style:
        table_width = get_term_width()

    if not header_defs:
        _header_defs = ColDefList.for_table([header_cols])
    elif isinstance(header_defs, ColDefList):
        _header_defs = header_defs
    else:
        _header_defs = ColDefList(header_defs)
    _header_defs.adjust_to_table([header_cols], table_width, style)

    if not col_defs:
        col_defs = _header_defs.copy()

    lazy_end = lazy_end and style.allow_lazy_header
    padding = 2 * style.cell_padding

    lines = []
    if style.top_border:
        line = str(style.header_top_line)
        delim = str(style.header_top_delimiter)
        left = str(style.header_top_left)
        right = line if lazy_end else str(style.header_top_right)
        border_lines = [line * (col.width + padding) for col in col_defs]
        border = delim.join(border_lines)
        border = left + border + right
        lines.append(border)

    headers = _get_table_row(
        values=header_cols,
        style=style,
        col_defs=_header_defs,
        table_width=table_width,
        lazy_end=lazy_end,
        is_header=True,
    )
    lines.append(headers)

    line = str(style.header_bottom_line)
    delim = str(style.header_bottom_delimiter)
    left = str(style.header_bottom_left)
    right = line if lazy_end else str(style.header_bottom_right)
    border_lines = []
    for col_idx in range(len(header_cols)):
        header_def = _header_defs[col_idx]
        col_def = None
        if col_idx < len(col_defs):
            col_def = col_defs[col_idx]
        if not style.align_char:
            h_line = line * (header_def.width + padding)
        else:
            h_line = line * header_def.width
            if col_def and col_def.align == "^":
                h_line = str(style.align_char) + h_line + str(style.align_char)
            elif col_def and col_def.align == ">":
                h_line = " " + h_line + str(style.align_char)
            else:
                h_line = " " + h_line + " "
        border_lines.append(h_line)
    border = delim.join(border_lines)
    border = left + border + right
    lines.append(border)

    return "\n".join(lines)


###############################################################################
# get_table
###############################################################################


def get_table(
    value_rows: List[List[str]],
    header_row: List[str] = None,
    style: TableStyle = NoBorderScreenStyle(),
    col_defs: List[str | ColDef] | ColDefList = None,
    header_defs: List[str | ColDef] | ColDefList = None,
    table_width: int = None,
    lazy_end: bool = False,
    separete_rows: bool = False,
) -> str:
    if not value_rows:
        return get_table(
            [["No data to display"]],
            style=style,
            table_width=table_width,
            lazy_end=lazy_end,
        )

    if not table_width and style.terminal_style:
        table_width = get_term_width()

    all_rows = value_rows.copy()
    if header_row:
        all_rows.insert(0, header_row.copy())

    if not col_defs:
        _col_defs = ColDefList.for_table(all_rows)
    elif isinstance(col_defs, ColDefList):
        _col_defs = col_defs
    else:
        _col_defs = ColDefList(col_defs)
    _col_defs.adjust_to_table(all_rows, table_width, style)

    if not header_row and style.force_header:
        header_row = [""] * len(_col_defs)

    # generate viable header definitions
    if header_row:
        real_header_defs = ColDefList()
        for col_def in _col_defs:
            real_header_defs.append(ColDef(width=col_def.width, align="^"))

        # if the defs are supplied, we only support alignment
        if header_defs:
            _header_defs = ColDefList(header_defs)
            for col_idx in range(len(_col_defs)):
                if col_idx < len(_header_defs):
                    header_def = _header_defs[col_idx]
                    if header_def.align:
                        real_header_defs[col_idx].align = header_def.align

    # Generate header and rows
    output_rows = []
    if header_row:
        row = get_table_header(
            header_cols=header_row,
            style=style,
            header_defs=real_header_defs,
            col_defs=_col_defs,
            table_width=table_width,
            lazy_end=lazy_end,
        )
        output_rows.append(row)
    else:
        if style.top_border:
            line = str(style.no_header_top_line)
            delim = str(style.no_header_top_delimiter)
            left = str(style.no_header_top_left)
            right = line if lazy_end else str(style.no_header_top_right)
            border_lines = [line * (col.width + 2) for col in _col_defs]
            border = delim.join(border_lines)
            border = left + border + right
            output_rows.append(border)

    rowcount = 0
    for values in value_rows:
        rowcount += 1
        lastrow = rowcount == len(value_rows)
        row = get_table_row(
            values=values,
            style=style,
            col_defs=_col_defs,
            table_width=table_width,
            lazy_end=lazy_end,
        )
        output_rows.append(row)
        if not lastrow and separete_rows and style.row_separator_line:
            sep_lines = []
            left = str(style.row_separator_left)
            right = line if lazy_end else str(style.row_separator_right)
            delim = str(style.row_separator_delimiter)
            padding = 2 * style.cell_padding
            for col_def in _col_defs:
                sep_line = str(style.row_separator_line) * (col_def.width + padding)
                sep_lines.append(sep_line)
            separator = delim.join(sep_lines)
            separator = left + separator + right
            output_rows.append(separator)

    if style.bottom_border:
        line = str(style.values_bottom_line)
        delim = str(style.values_bottom_delimiter)
        left = str(style.values_bottom_left)
        right = line if lazy_end else str(style.values_bottom_right)
        border_lines = [line * (col.width + 2) for col in _col_defs]
        border = delim.join(border_lines)
        border = left + border + right
        output_rows.append(border)

    return "\n".join(output_rows)



# Query Property Constants

# collection
ALL_DATABASES = 'databases'
ALL_QUERIES = 'queries'

# per query
DATABASE_NAME = 'dbname'
SOLUTION = 'solution'

# parsed
EMPTY = 'empty'
SQL_VERB = 'verb'
SELECT_EXPRESSIONS = 'expressions'
HAS_COUNT_STAR = "hasStar"
SELECT_HAS_AS = 'selectHasAs'
FROM_HAS_AS = 'fromHasAs'
EXPR_NAME = "name"
EXPR_ACTUAL = 'actual'
EXPR_HAS_ALIAS = 'hasAlias'
TABLES = "tables"
TABLE_NAME = 'name'
TABLE_ALIAS = 'alias'
HAS_AS = 'hasAs'
MODIFIED_TABLE = 'modifiedTable'

# execution results

# DQL
ROW_COUNT = 'rowCount'
COLUMN_COUNT = 'colCount'
COLUMN_NAMES = 'colNames'
FIRST_ROW = 'firstRow'
LAST_ROW = 'lastRow'

# DML
TABLE_NAME = "name"
ROWS_AFFECTED = 'rowsAffected'

# Query Property Collections
SANITY_PROPS = (
    SQL_VERB,
    DATABASE_NAME,
    ROW_COUNT,
    COLUMN_COUNT,
    ROWS_AFFECTED,
    MODIFIED_TABLE
)

FULL_PROPS = (
    SQL_VERB,
    DATABASE_NAME,
    HAS_COUNT_STAR,
    SELECT_HAS_AS,
    FROM_HAS_AS,
    # DQL
    ROW_COUNT,
    COLUMN_COUNT,
    # COLUMN_NAMES,
    FIRST_ROW,
    LAST_ROW,
    # DML
    ROWS_AFFECTED,
    MODIFIED_TABLE,
)

__propToFriendlyDict__ = {
    SQL_VERB: 'sql verb',
    DATABASE_NAME: 'database name',
    HAS_COUNT_STAR: "uses COUNT(*) to count records",
    SELECT_HAS_AS: 'use AS keyword for all aliases in SELECT clause',
    FROM_HAS_AS: 'use AS keyword for all aliases in FROM clause',
    ROW_COUNT: 'number of rows',
    COLUMN_COUNT: 'number of columns',
    ROWS_AFFECTED: 'rows affected',
    MODIFIED_TABLE: 'table name',
    FIRST_ROW: 'first row',
    LAST_ROW: 'last row'
}

def friendly(propName):
    friendlier = propName.lower().replace('_',' ')
    friendly = __propToFriendlyDict__.setdefault(propName, friendlier)
    return friendly

# Filenames
SANITY_CHECK_FILENAME = 'sanitycheck.json'
FULL_CHECK_FILENAME = 'fullcheck.json'
FEEDBACK_FILENAME = 'FEEDBACK.md'
FEEDBACK_DATA_FILENAME = '.sanity-check/raw-feedback.json'

TEMPLATE_DIRECTORY = '_template_'
SOLUTION_DIRECTORY = '_solution_'



#!/usr/bin/python3

# GENERATE FEEDBACK
# 
#     This script generates a FEEDBACK.md file that contains results of the
#     full suite of automated tests run against all sql files in this
#     fullcheck.json file.
# 
#     TODO:
#       * Add UsesDistinct
#         - And catch DISTINCT(...)
#       * Add UsesLimit
#       * Add UsesLike (or REGEXP)
#       * Add UsesOuterJoin
#       * Add basic style checks / linting
#       * DONE Support multi-line/fully-delimited comments /* ... */
#       * ORDER BY "stringLiteral"
#       * Add explicit support for ORDER BY
#       * Suppress "expect to find value" feedback for missing cols
#       

from commonutil import ps, wb, fullpath
from commonsql import resetDatabase, parseSQL, getProperties
import props
import json
from os import getcwd
from os.path import isfile
from pathlib import(Path)
import re
from pprint import pformat
from collections import Counter

import logging
logging.basicConfig()
baseLoggingLevel = logging.INFO

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)

_debugQueries = []

def lower(string):
    if string:
        string = string.strip().lower()
    return string

def clean(string):
    if string:
        string = string.strip().lower().replace(' ','').replace('_','')
    return string

def get_col_name(col):
    name = col[props.EXPR_NAME]
    actual = col[props.EXPR_ACTUAL]
    # has_alias = col[props.EXPR_HAS_ALIAS]
    if name:
        return name
    else:
        return actual

def get_col_alias(col):
    name = col[props.EXPR_NAME]
    has_alias = col[props.EXPR_HAS_ALIAS]
    if has_alias:
        return name
    else:
        return None

func_pattern = re.compile(r'(?P<name>\w+)\s*\((?P<args>.*)\)$')
def get_col_table_field(col, aliases):
    actual = col[props.EXPR_ACTUAL]
    table = None
    field = None
    match = func_pattern.match(actual)
    if not match and "." in actual:
        parts = actual.split(".")
        # table = lower(parts[0])
        table = parts[0]
        if table in aliases:
            table = aliases[table]
        # field = lower(parts[1])
        field = parts[1]
    return table, field

def func(col):
    actual = col[props.EXPR_ACTUAL]
    name = None
    match = func_pattern.match(actual)
    if match:
        name = lower(match.group('name'))    
    return name

dirname = Path(getcwd()).name


with open(fullpath(props.FULL_CHECK_FILENAME)) as fi:
    query_data = json.load(fi)

databases = query_data[props.ALL_DATABASES]

for dbname, dbfilename in databases.items():
    logger.debug(f"Resetting {dbname}...")
    file = fullpath(dbfilename)
    resetDatabase(dbname, file)

logger.debug(f"Executing tests...")

queries = query_data[props.ALL_QUERIES]
num_queries = 0
raw_feedback = []
query_feedback = []

class StopException(Exception):
    pass

for query_name, expected_props in queries.items():

    debug_query = query_name in _debugQueries
    if debug_query:
        logging.getLogger().setLevel(logging.DEBUG)
        logger.debug(f"Forcing DEBUG level on {query_name}")
    else:
        logging.getLogger().setLevel(baseLoggingLevel)

    is_select = expected_props.get(props.SQL_VERB) == 'select'

    raw_issues = []
    issues = []
    info = []
    stu_props = dict(expected_props)

    if '#' not in query_name:
        num_queries += 1

    # Outer TRY block to allow "continuing" the loop while still adding issues
    # at the end
    try:

        sql = ''

        try:
            with open(query_name + '.sql') as fi:
                sql = fi.read()
        except Exception as ex:
            issues.append(f"Error reading SQL file: {ex}")
            raise StopException()
        
        try:
            sql = parseSQL(sql, stu_props)
            logger.debug(sql)
        except Exception as ex:
            issues.append(f"Error parsing SQL: {ex}")
            logger.debug(f"Error parsing SQL: {ex}")
            raise StopException()

        if len(sql) == 0:
            verb = stu_props.get(props.SQL_VERB)
            if verb:
                raw_issues.append({'issue': props.SQL_VERB})
                issues.append(f"Unknown SQL verb {verb.upper()}")
                logger.debug(f"Unknown SQL verb {verb.upper()}")
            else:
                raw_issues.append({'issue': props.EMPTY})
                issues.append("EMPTY")
                logger.debug("EMPTY")

        else:

            try:
                # copy expectedprops to fill in
                logger.debug("expected_props: " + pformat(expected_props))
                getProperties(sql, stu_props)
                logger.debug("stu_props: " + pformat(stu_props))
            except Exception as ex:
                issues.append(f"Error executing SQL: {ex}")
                logger.debug(f"Error executing SQL: {ex}")
                raise StopException()
        
            if is_select:
                # create a column map to try handle cases where students add aliases
                # when not needed / wanted or aliases with/without spaces.

                exp_col_names = expected_props.get(props.COLUMN_NAMES)
                exp_cols = expected_props.get(props.SELECT_EXPRESSIONS)
                exp_tables = expected_props.get(props.TABLES)
                exp_table_names = [lower(t[props.TABLE_NAME]) for t in exp_tables]
                exp_table_dups = {k: 0 for k, v in Counter(exp_table_names).items() if v > 1}
                exp_aliases = {}
                for table in exp_tables:
                    if props.TABLE_ALIAS in table:
                        name = table[props.TABLE_NAME]
                        alias = table[props.TABLE_ALIAS]
                        if name in exp_table_dups:
                            counter = exp_table_dups[name]
                            exp_table_dups[name] = counter + 1
                            name = name + "_" + str(counter)
                        exp_aliases[alias] = name
                
                stu_cols = stu_props.get(props.SELECT_EXPRESSIONS)
                stu_tables = stu_props.get(props.TABLES)
                stu_table_names = [lower(t[props.TABLE_NAME]) for t in stu_tables]
                stu_table_dups = {k: 0 for k, v in Counter(stu_table_names).items() if v > 1}
                stu_aliases = {}
                for table in stu_tables:
                    if props.TABLE_ALIAS in table:
                        name = table[props.TABLE_NAME]
                        alias = table[props.TABLE_ALIAS]
                        if name in stu_table_dups:
                            counter = stu_table_dups[name]
                            stu_table_dups[name] = counter + 1
                            name = name + "_" + str(counter)
                        stu_aliases[alias] = name

                logger.debug('EXPECTED ALIASES:\n' + pformat(exp_aliases))
                logger.debug('STUDENT ALIASES:\n' + pformat(stu_aliases))
                
                col_map = []

                # FIRST PASS (exact and high confidence matches)

                logger.debug(f"TESTING {query_name}")
                
                for i in range(len(exp_cols)):
                    col_map.append(None)

                    exp_col = exp_cols[i]
                    logger.debug(f"exp_col: {pformat(exp_col)}")

                    exp_col_name = get_col_name(exp_col)
                    logger.debug(f"exp_col_name: {exp_col_name}")

                    exp_col_alias = get_col_alias(exp_col)
                    logger.debug(f"exp_col_alias: {exp_col_alias}")

                    exp_col_clean = clean(exp_col[props.EXPR_NAME])
                    exp_col_actual = exp_col[props.EXPR_ACTUAL]
                    exp_col_has_alias = exp_col[props.EXPR_HAS_ALIAS]

                    exp_col_table, exp_col_field = get_col_table_field(exp_col, exp_aliases)

                    logger.debug(f"EXPECTED: {exp_col_table}.{exp_col_field}")


                    for j in range(len(stu_cols)):
                        if j in col_map:
                            continue

                        stu_col = stu_cols[j]
                        stu_col_name = get_col_name(stu_col)
                        stu_col_alias = get_col_alias(stu_col)
                        stu_col_clean = clean(stu_col[props.EXPR_NAME])
                        stu_col_actual = stu_col[props.EXPR_ACTUAL]
                        stu_col_has_alias = stu_col[props.EXPR_HAS_ALIAS]

                        stu_col_table, stu_col_field = get_col_table_field(stu_col, stu_aliases)

                        logger.debug(f"STUDENT: {stu_col_table}.{stu_col_field}")

                        # First, if both queries specify aliases and they are
                        # similar, call it a match
                        if exp_col_alias and stu_col_alias and clean(exp_col_alias) == clean(stu_col_alias):
                            col_map[i] = j
                            if lower(exp_col_alias) == lower(stu_col_alias):
                                logger.debug(f'+ EXACT ALIAS: {exp_col_alias} == {stu_col_alias}')
                            else:
                                logger.debug(f'+ CLEAN ALIAS: {exp_col_alias} == {stu_col_alias}')
                                info.append(f'Unexpected alias {stu_col_alias} found for {exp_col_alias}')
                                raw_issues.append({'issue': props.COLUMN_NAMES, 'val': exp_col_alias})
                            break

                        # Next, if table.field notation is used, we can be
                        # certain to choose the right column
                        if exp_col_field and stu_col_field:
                            if exp_col_field == stu_col_field and exp_col_table == stu_col_table:
                                logger.debug(f'- FIELD MATCH: {exp_col_table}.{exp_col_field} == {stu_col_table}.{stu_col_field} ({exp_col_name})')
                                col_map[i] = j
                                
                                # Now compare names to get alias comparison as needed
                                if lower(exp_col_name) != lower(stu_col_name):
                                    if stu_col_has_alias:
                                        info.append(f'Unexpected alias {stu_col_name} found for {exp_col_name}')
                                    else:
                                        issues.append(f'Missing alias {exp_col_name} for {stu_col_name}')
                                    raw_issues.append({'issue': props.COLUMN_NAMES, 'val': exp_col_name})
                                break
                            else:
                                continue
                                                    # if the col names match, then just assume we've found a
                        # Find exact name matches
                        if exp_col_name and lower(exp_col_name) == lower(stu_col_name):
                            logger.debug(f'+ EXACT COLUMN: {exp_col_name} == {stu_col_name}')
                            col_map[i] = j
                            break

                        # if the col names match after removing spaces and
                        # underscores, then also assume a match (likely an alias
                        # added that matches the column name)
                        if exp_col_clean and exp_col_clean == stu_col_clean:
                            logger.debug(f'- CLEAN COLUMN: {exp_col_clean} == {stu_col_clean}')
                            col_map[i] = j
                            info.append(f'Unexpected alias {stu_col_name} found for {exp_col_name}')
                            raw_issues.append({'issue': props.COLUMN_NAMES, 'val': exp_col_name})
                            break

                        # if the actual expressions match, we've found a match
                        if exp_col_actual and clean(exp_col_actual) == clean(stu_col_actual):
                            logger.debug(f'+ EXACT: {exp_col_actual} == {stu_col_actual}')
                            col_map[i] = j
                            if stu_col_has_alias:
                                info.append(f'Unexpected alias {stu_col_name} found for {exp_col_actual}')
                                raw_issues.append({'issue': props.COLUMN_NAMES, 'val': exp_col_name})
                            elif exp_col_has_alias:
                                issues.append(f'Missing alias {exp_col_name} for {stu_col_actual}')
                                raw_issues.append({'issue': props.COLUMN_NAMES, 'val': exp_col_name})
                            break

                        # One is dotted the other is not...fuzzier but still confident
                        if exp_col_field or stu_col_field:
                            if exp_col_field == lower(stu_col_actual) or stu_col_field == lower(exp_col_actual):
                                logger.debug(f'- MISSING ALIAS: {stu_col_clean} == {exp_col_field} ({exp_col_name})')
                                col_map[i] = j
                                if stu_col_has_alias:
                                    info.append(f'Unexpected alias {stu_col_name} found for {exp_col_actual}')
                                    raw_issues.append({'issue': props.COLUMN_NAMES, 'val': exp_col_name})
                                elif exp_col_has_alias:
                                    issues.append(f'Missing alias {exp_col_name} for {stu_col_actual}')
                                    raw_issues.append({'issue': props.COLUMN_NAMES, 'val': exp_col_name})
                                break

                # SECOND PASS (try our best)

                for i in range(len(exp_cols)):
                    if col_map[i] != None:
                        continue

                    exp_col = exp_cols[i]
                    exp_col_name = get_col_name(exp_col)
                    sep_col_clean = clean(exp_col[props.EXPR_NAME])
                    exp_col_actual = exp_col[props.EXPR_ACTUAL]
                    exp_col_has_alias = exp_col[props.EXPR_HAS_ALIAS]

                    exp_col_func = None
                    
                    match = func_pattern.match(exp_col_actual)
                    if match:
                        exp_col_func = lower(match.group('name'))

                    for j in range(len(stu_cols)):
                        if j in col_map:
                            continue

                        stu_col = stu_cols[j]
                        stu_col_name = get_col_name(stu_col)
                        stu_col_clean = clean(stu_col[props.EXPR_NAME])
                        stu_col_actual = stu_col[props.EXPR_ACTUAL]
                        stu_col_has_alias = stu_col[props.EXPR_HAS_ALIAS]

                        stu_col_func = None

                        match = func_pattern.match(stu_col_actual)
                        if match:
                            stu_col_func = lower(match.group('name'))

                        # if the funcs match, then we'll also assume a match
                        logger.debug(exp_col_func, stu_col_func)
                        if exp_col_func and exp_col_func == stu_col_func:
                            logger.debug(f'- FUNC MATCH: {exp_col_func} = {stu_col_func} ({exp_col_name})')
                            col_map[i] = j
                            if stu_col_has_alias:
                                info.append(f'Unexpected alias {stu_col_name} found for {exp_col_actual}')
                                raw_issues.append({'issue': props.COLUMN_NAMES, 'val': exp_col_name})
                            elif exp_col_has_alias:
                                issues.append(f'Missing alias {exp_col_name} for {stu_col_actual}')
                                raw_issues.append({'issue': props.COLUMN_NAMES, 'val': exp_col_name})
                            break

                    # At this point, we give up
                    if col_map[i] == None:
                        issues.append(f"Expected to find a column named {exp_col_name}")
                        raw_issues.append({'issue': props.COLUMN_NAMES, 'val': exp_col_name})

                logger.debug("col_map: " + pformat(col_map))
            
            for prop_name, exp_prop in expected_props.items():
                if prop_name not in props.FULL_PROPS:
                    continue
                
                exp_prop = expected_props[prop_name]

                stu_prop = stu_props[prop_name]
                friendly = props.friendly(prop_name)
                
                if prop_name in (props.SELECT_HAS_AS, props.FROM_HAS_AS):
                    if stu_prop == False:
                        raw_issues.append({'issue': prop_name})
                        issues.append(friendly)
                elif prop_name in (props.FIRST_ROW, props.LAST_ROW):
                    for i in range(len(exp_prop)):
                        exp_val = exp_prop[i]
                        col_name = exp_col_names[i]
                        if col_map[i] != None:
                            stu_index = col_map[i]
                            stu_val = stu_prop[stu_index]
                            if exp_val.lower() == stu_val.lower():
                                continue
                        raw_issues.append({'issue': friendly, 'val': col_name})
                        issues.append(f"Expected to find {exp_val} in {col_name} for {friendly}")
                elif type(exp_prop) is str:
                    if exp_prop.lower() != stu_prop.lower():
                        raw_issues.append({'issue': friendly, 'col': stu_prop})
                        issues.append(f"Expected {exp_prop} for {friendly} but retrieved {stu_prop}")
                else:
                    if exp_prop != stu_prop:
                        raw_issues.append({'issue': friendly, 'col': stu_prop})
                        issues.append(f"Expected {exp_prop} for {friendly} but retrieved {stu_prop}")
    except StopException:
        pass
    except Exception as ex:
        issues.append(f"Unexpected Exception: {ex}")
        logging.exception("Unexpected exception in ridiculously gigantic try block")

    if raw_issues:
        raw_feedback.append({
            'query_name': query_name,
            'issues': raw_issues,
            'sql': sql
        })

    query_feedback.append({
        'query_name': query_name,
        'issues': issues,
        'info': info,
        'student_sql': sql,
        'solution_sql': expected_props[props.SOLUTION]
    })

logger.debug("query_feedback: " + pformat(query_feedback))

logger.debug(f'...writing issues to {props.FEEDBACK_DATA_FILENAME}')
with open(props.FEEDBACK_DATA_FILENAME, 'w') as fi:
    fi.write(json.dumps(raw_feedback, indent=2))

if isfile(props.FEEDBACK_FILENAME):
    logger.info(f"SKIPPING: {dirname} because {props.FEEDBACK_FILENAME} already exists")
    exit(0)

points_per_query = round(100 / num_queries)

logger.debug(f'...writing issues to {props.FEEDBACK_FILENAME}')
with open(props.FEEDBACK_FILENAME, 'w') as fi:

    fi.write("# FEEDBACK\n")
    fi.write("\n")

    for query in query_feedback:
        fi.write('-'*78 + '\n')
        query_name = query['query_name']
        challenge_query = '#' in query_name
        info = query['info']
        issues = query['issues']

        if len(issues) == 0:
            fi.write(f"### ✓  {query_name}\n")
            wb(info, fi)
        else:
            if issues[0] == "EMPTY":
                line = f"### ✗  {query_name}"
                if not challenge_query:
                    line += f" (DEDUCTION: {points_per_query})"
                fi.write(line + "\n")
                wb(["NO QUERY PROVIDED"], fi)
            else:
                student_sql = query['student_sql']
                solution_sql = query['solution_sql']
                line = f"### ✗  {query_name}"
                if not challenge_query:
                    line += f" (DEDUCTION: )"
                fi.write(line + "\n")
                wb(info, fi)
                wb(issues, fi)
                fi.write("\n**YOUR SQL**\n")
                fi.write("\n```\n")
                fi.write(student_sql)
                fi.write("\n```\n")
                fi.write("\n**SOLUTION SQL**\n")
                fi.write("\n```\n")
                fi.write(solution_sql)
                fi.write("\n```\n\n")

logger.info(f"{len(queries)} checked for {dirname}")



#!/usr/bin/python3

# GENERATE FEEDBACK SCORE
# 
#   This parses FEEDBACK.md for deductions. It sums these deductions and applies
#   a final score to the file.

import logging
import props
from os import getcwd
from os.path import isfile

logging.basicConfig(level=logging.INFO)

if not isfile(props.FEEDBACK_FILENAME):
    logging.info(f"skipping {getcwd()} because {props.FEEDBACK_FILENAME} doesn't exist")
    exit(0)

logging.info(f"processing {getcwd()}...")

deduction_const = '(DEDUCTION:'
total_deductions = 0.0
deduction_count = 0
nonempty_deductions = 0

with open(props.FEEDBACK_FILENAME, 'r') as fi:
    feedback_lines = fi.readlines()

for line in feedback_lines:
    line = line.strip()

    if deduction_const in line:
        deduction_count += 1

        idx = line.index(deduction_const) + len(deduction_const)

        ded_str = ''
        for c in line[idx:]:
            if c == '.' or c.isdigit():
                ded_str += c
            elif c == ')':
                break

        if not ded_str:
            continue

        try:
            ded = float(ded_str)
            total_deductions += ded
            nonempty_deductions += 1
        except:
            print(f"Unable to parse deduction for '{line}'")
            continue

score = max( 100 - total_deductions, 0 ) # 0 is the lowest score possible

logging.debug(f"{deduction_count} deductions found")
logging.debug(f"{nonempty_deductions} non-zero deductions found")
logging.debug(f"{total_deductions} total deductions")

if deduction_count == nonempty_deductions:
    logging.info(f"SCORE: {score}")

    score_prefix = '## SCORE: '
    score_line = f'{score_prefix}{score:g}\n'
    if feedback_lines[1].startswith(score_prefix):
        feedback_lines[1] = score_line
    else:
        feedback_lines.insert(1, score_line)

    with open(props.FEEDBACK_FILENAME, 'w') as fi:
        fi.writelines(feedback_lines)

else:
    logging.info(f"SKIPPING DUE TO EMPTY SCORES")


from .drawing import (
    BoxChars,
    get_line,
    get_single_line,
    get_double_line,
)

from .terminal import (
    get_choice,
    select_choice,
    select_file,
    select_dir,
    get_term_width,
    draw_line,
    draw_single_line,
    draw_double_line,
    strip_ansi,
)

from .csv import dict_from_csv

# Access others via files



import mariadb as db
import props
from os import system
import regex as re
from sqlglot import parse_one, exp
from pprint import pformat as pf

import logging
logger = logging.getLogger(__name__)

# Reset a database using a SQL script
def resetDatabase(dbname, filename):
    logger.info(f"Dropping {dbname}")
    system(f'mysql -u root -e "DROP DATABASE IF EXISTS {dbname}"')
    logger.info(f"Sourcing {filename}")
    system(f'mysql -u root -e "SOURCE {filename}"')

# sl_comment_pattern = re.compile(r'(\/\/|--).*')
full_line_comment_pattern = re.compile(r'^\s*(#|--|\/\/).*', re.MULTILINE)
end_of_line_comment_pattern = re.compile(r'(#|--|\/\/)[^"\n\r]*(?:"[^"\n\r]*"[^"\n\r]*)*[\r\n]', re.MULTILINE)
multiline_comment_pattern = re.compile(r'\/\*.*?(\*\/|$)', re.DOTALL)

# Remove comments and blank lines from SQL
def removeComments(sql):
    sql = multiline_comment_pattern.sub('', sql)
    sql = full_line_comment_pattern.sub('', sql)
    logger.debug('Full Line:\n' + sql)
    sql = end_of_line_comment_pattern.sub('', sql)
    logger.debug('End of Line:\n' + sql)
    return sql

# Pull (potentially) multiple queries from a file. Though incorrect, be
# resilient to queries that have blank lines in between, as the MySQL plugin
# treats line-separated queries as separate.
def getQueries(sql):
    sql = removeComments(sql)
    queries = []
    sql_lines = []
    blank_line_count = 0
    for line in sql.splitlines():
        end = False
        line = line.rstrip()
        strip_line = line.lstrip()
        if len(strip_line) > 0:
            blank_line_count = 0
            if line.endswith(';'):
                line = line[:-1]
                end = True
            sql_lines.append(line)
        else:
            # tolerate a single blank line before assuming a new query
            if len(sql_lines) > 2 and blank_line_count > 0:
                end = True
            else:
                blank_line_count += 1
        
        # Dump the query lines
        if end:
            queries.append('\n'.join(sql_lines))
            sql_lines.clear()
    
    # Get the last query if not semi-colon or blank-line terminated
    if sql_lines:
        queries.append('\n'.join(sql_lines))

    return queries


# Fully parse a SELECT query to retrieve pertinent properties
def parseSelect(sql, properties):
    glot = parse_one(sql,read="mysql")
    
    # these determine a value for the whole query
    select_has_count_star = False
    select_has_as = None
    from_has_as = None

    # process the SELECT clause
    expressions = []
    for expr in glot.expressions:
        expression = {}
        expression[props.EXPR_NAME] = expr.alias_or_name
        expression[props.EXPR_HAS_ALIAS] = False

        if type(expr) is exp.Alias:
            expression[props.EXPR_HAS_ALIAS] = True
            if ' as ' in expr.sql().lower():
                expression[props.HAS_AS] = True
                select_has_as = select_has_as is None or select_has_as and True
            else:
                expression[props.HAS_AS] = False
                select_has_as = False
            expr = expr.unalias()

        expression[props.EXPR_ACTUAL] = expr.sql()

        for node in expr.walk():
            if type(node) is exp.Count:
                if node.this.is_star:
                    expression[props.HAS_COUNT_STAR] = select_has_count_star = True
                else:
                    expression[props.HAS_COUNT_STAR] = False

        expressions.append(expression)

    # process the FROM clause
    tables = []
    for tab in glot.find_all(exp.Table):
        table = {}
        table[props.TABLE_NAME] = tab.name
        if tab.alias:
            table[props.TABLE_ALIAS] = tab.alias
            if ' as ' in tab.sql().lower():
                table[props.HAS_AS] = True
                from_has_as = from_has_as is None or from_has_as and True
            else:
                table[props.HAS_AS] = False
                from_has_as = False
        
        tables.append(table)
    
    properties[props.SELECT_EXPRESSIONS] = expressions
    properties[props.TABLES] = tables
    properties[props.HAS_COUNT_STAR] = select_has_count_star
    properties[props.SELECT_HAS_AS] = select_has_as
    properties[props.FROM_HAS_AS] = from_has_as


# Return the _first_ SELECT/INSERT/UPDATE/DELETE query. As appropriate, also
# fills the database name, verb, and table properties as well as all "parsed"
# properties of a SELECT query.
def parseSQL(sql, properties = {}):
    properties[props.SQL_VERB] = ''

    sql_queries = getQueries(sql)

    for query_sql in sql_queries:
        words = query_sql.split()
        sql_verb = words[0].strip().lower()

        properties[props.SQL_VERB] = sql_verb

        match sql_verb:
            case 'use':
                properties[props.DATABASE_NAME] = words[1].strip().lower()
                continue
            case 'insert':
                # because some don't add space after parens...
                word3 = words[2]
                more_words = word3.split('(')
                properties[props.MODIFIED_TABLE] = more_words[0]
            case 'update':
                properties[props.MODIFIED_TABLE] = words[1]
            case 'delete':
                properties[props.MODIFIED_TABLE] = words[2]
            case 'select':
                parseSelect(query_sql, properties)
        
        return query_sql
    return ''

# Retrieve a connection to the named database.
connections = {}
def getConnection(dbname=None):
    global connections

    if not dbname:
        return db.connect(
                host="127.0.0.1",
                port=3306,
                user='root',
                connect_timeout=1
            )

    if not dbname in connections.keys():
        connections[dbname] = db.connect(
                    host="127.0.0.1",
                    port=3306,
                    user='root',
                    database=dbname,
                    connect_timeout=1
        )
    return connections[dbname]

# Convert each item in a tuple (row) to a string so that the silly
# json.dumps() function works.
def stringifyRow(row):
    result = list(row)
    for i in range(len(result)):
        result[i] = str(result[i])
    return result

# Retrieve the properties of a given query.
def getProperties(sql, properties):
    dbname = properties[props.DATABASE_NAME]
    conn = getConnection(dbname)

    with conn.cursor() as cursor:
        cursor.execute(sql)
        try:
            data = cursor.fetchall() # throws if query doesn't have results (isn't DQL)

            # sanity props
            properties[props.ROW_COUNT] = len(data)
            properties[props.COLUMN_COUNT] = len(cursor.description)

            #full props
            properties[props.COLUMN_NAMES] = [item[0] for item in cursor.description]
            if len(data) > 0:
                properties[props.FIRST_ROW] = stringifyRow(data[0])
            if len(data) > 1:
                properties[props.LAST_ROW] = stringifyRow(data[-1])
        except:
            # assume query wasn't supposed to return rows and is thus DML
            properties['rowsAffected'] = cursor.rowcount
            conn.commit()



from typing import Dict
from dataclasses import dataclass

from f_table import MarkdownStyle, NoBorderScreenStyle, get_table, get_table_row

from au.common.datetime import get_friendly_local_datetime

from .pytest_data import Results


@dataclass
class ScoringParams:
    max_score: float = 10
    pytest_weight: float = 1
    pylint_weight: float = 0


@dataclass
class Score:
    pytest_pct: float
    pylint_pct: float
    overall_score: float


def get_student_scores(
    student_results: Dict, scoring_params: ScoringParams = ScoringParams()
) -> Score:
    # Calculate score
    pt_pct = student_results.setdefault("pytest_pct", 0)
    pl_pct = student_results.setdefault("pylint_pct", 0)
    total_score = None
    if scoring_params:
        max = scoring_params.max_score
        pt_weight = scoring_params.pytest_weight
        pl_weight = scoring_params.pylint_weight
        total_score = round(max * ((pt_pct * pt_weight) + (pl_pct * pl_weight)), 2)

    return Score(pt_pct, pl_pct, total_score)


def get_summary(
    student_results: Dict, scoring_params: ScoringParams = None, get_markdown=False
) -> str:
    summary_values = [
        ["Student Name", student_results["name"]],
        [
            "Last Commit Date",
            get_friendly_local_datetime(student_results["commit_date"]),
        ],
        ["Last Commit Author", student_results["commiter_name"]],
        ["Last Commit Message", student_results["commit_message"]],
        ["Total Commit Count", student_results["num_commits"]],
    ]

    pytest_weight = ""
    pylint_weight = ""
    final_score_row = None
    if scoring_params:
        pytest_weight = f"(weight: {scoring_params.pytest_weight*100:.4g}%)"
        pylint_weight = f"(weight: {scoring_params.pylint_weight*100:.4g}%)"
        scores = get_student_scores(student_results, scoring_params)
        final_score_row = [
            "Calculated Score",
            f"{scores.overall_score:g} / {scoring_params.max_score:g}",
        ]
    else:
        scores = get_student_scores(student_results, None)

    summary_values.append(
        ["Functionality Score", f"{scores.pytest_pct*100:.4g}% {pytest_weight}"],
    )

    try:
        pytest_results = Results.from_dict(student_results.get("pytest_results"))
        for test_class in pytest_results.test_classes.values():
            tot = len(test_class.tests)
            correct = test_class.pass_count
            summary_values.append([f" + {test_class.name}", f"{correct} out of {tot}"])
    except:
        pass

    summary_values.append(
        ["Code Style Score", f"{scores.pylint_pct*100:.4g}% {pylint_weight}"],
    )

    if final_score_row:
        summary_values.append(final_score_row)

    past_due = student_results.setdefault("past_due", None)
    if past_due:
        summary_values.append(["Past Due", past_due])

    style = MarkdownStyle() if get_markdown else NoBorderScreenStyle()
    return get_table(summary_values, col_defs=["25T", "52"], lazy_end=True, style=style)


def get_summary_row(label: str, value: str):
    return get_table_row(
        [label, value], col_defs=["25T", "52"], lazy_end=True, style=MarkdownStyle()
    )



"""Contains the GitRepo class and related utility functions."""

from dataclasses import dataclass
from typing import Iterable, List
import sys
from datetime import datetime
import os
from textwrap import dedent
import logging
import subprocess
from pathlib import Path

logger = logging.getLogger(__name__)


class NotARepoError(ValueError):
    """Indicates that a specified directory is not a Git repository"""


class GitCommandError(Exception):
    """Indicates that a git command exited with a non-zero exit code"""


@dataclass
class Commit:
    """Basic dataclass describing a single Git commit."""

    hash: str
    message: str
    date: datetime
    author_name: str
    author_email: str
    committer_name: str


class GitRepo:
    """Basic wrapper around git for when GitPython is overkill."""

    def __init__(self, path: Path | str = ".", traverse_parents: bool = False):
        path = Path(path).resolve()
        base = GitRepo.get_repository_root(path)
        if base:
            if traverse_parents or base == path:
                self.root_dir = path
                self.name = path.name
                return
        root = " root" if not traverse_parents else ""
        raise NotARepoError(f"{path} is not a valid repository{root}")

    def __str__(self):
        return self.name

    @staticmethod
    def git(*args, path: Path | str = Path.cwd()) -> subprocess.CompletedProcess:
        """Execute a single git command in the provided path."""
        path = Path(path)
        cmd = ["git", "--no-pager"] + list(args)
        prev_cwd = os.getcwd()
        try:
            os.chdir(path)
            return subprocess.run(cmd, capture_output=True, text=True, check=True)
        except FileNotFoundError:
            logger.exception(
                "Error trying to run `git` command. Git may not be installed."
            )
        except KeyboardInterrupt:
            sys.exit(1)
        except subprocess.CalledProcessError as ex:
            err = f"""
                COMMAND:    {ex.cmd}
                CODE:       {ex.returncode}
                OUTPUT:     {ex.output}
                STDERR:     {ex.stderr}
                """
            logger.error(dedent(err).strip())
            raise GitCommandError() from ex
        finally:
            os.chdir(prev_cwd)
        return None

    @staticmethod
    def get_repository_root(path: Path | str) -> Path:
        """
        Gets the base directory of a Git repository at `path` or None if it is
        not a repository.
        """
        path = Path(path).resolve()
        if not path.is_dir():
            raise NotADirectoryError(f"{path} is not a directory")
        try:
            proc = GitRepo.git("rev-parse", "--show-toplevel", path=path)
            root = Path(proc.stdout.strip())
            if root.is_dir() and path.exists():
                return root.resolve()
        except GitCommandError:
            pass
        return None

    @staticmethod
    def is_repository(path: Path | str) -> bool:
        """Determines whether `path` is a Git repository."""
        path = Path(path).resolve()
        if not path.is_dir():
            raise NotADirectoryError(f"{path} is not a directory")
        return GitRepo.get_repository_root(path) is not None

    @staticmethod
    def is_repository_root(path: Path | str) -> bool:
        """Determines whether qpath` is at the root of a Git repository."""
        path = Path(path).resolve()
        if not path.is_dir():
            raise NotADirectoryError(f"{path} is not a directory")
        root = GitRepo.get_repository_root(path)
        return root and root == path

    @staticmethod
    def clone(source_url: str, dest_dir: Path | str) -> None:
        """
        Clone a git repository from a source URL into a subdirectory `dest_dir`.
        If `dest_dir` does not exist, it will be created. If `dest_dir` is not
        empty, the command will fail.
        """
        dest_dir = Path(dest_dir).resolve()
        if dest_dir.exists():
            for _ in dest_dir.iterdir():
                raise FileExistsError(f"{dest_dir} is not empty")
        GitRepo.git("clone", source_url, str(dest_dir))

    @staticmethod
    def get_user_name() -> str:
        """Retrieve the current user's name from the global git config."""
        try:
            proc = GitRepo.git("config", "--global", "--get", "user.name")
            return proc.stdout.strip()
        except GitCommandError:
            return None

    @staticmethod
    def get_user_email() -> str:
        """Retrieve the current user's email from the global git config."""
        try:
            proc = GitRepo.git("config", "--global", "--get", "user.email")
            return proc.stdout.strip()
        except GitCommandError:
            return None

    def _git(self, *args) -> subprocess.CompletedProcess:
        return GitRepo.git(*args, path=self.root_dir)

    def is_dirty(self, untracked_files: bool = True) -> bool:
        """Determine if a Git repository contains changed or untracked files."""
        if untracked_files:
            proc = self._git("status", "-s")
        else:
            proc = self._git("status", "-s", "-u", "no")
        output = proc.stdout.strip() + proc.stderr.strip()
        if output:
            return True
        else:
            return False

    def needs_pull(self) -> bool:
        """Check whether remote is ahead of local."""
        proc = self._git("fetch", "--dry-run")
        output = proc.stdout.strip() + proc.stderr.strip()
        if output:
            return True
        else:
            return False

    def pull(self) -> None:
        """Pull changes from the default Git repository remote."""
        proc = self._git("fetch")
        output = proc.stdout.strip() + proc.stderr.strip()
        if not output:
            return False
        self._git("merge")
        return True

    def add(self, files: Iterable[Path | str] = None) -> None:
        """Run `git add` to stage files. If files is empty, add `.` (all)."""
        args = []
        if files:
            for file in files:
                file = Path(file)
                if file.exists() and file.is_relative_to(self.root_dir):
                    files.append(file.relative_to(self.root_dir))
        if not args:
            args.append(".")
        self._git("add", *args)

    def commit(self, message: str) -> None:
        """Commit any staged changes using the provided message."""
        if not message:
            raise ValueError("`messageq` cannot be empty.")
        self._git("commit", "-m", message)

    def push(self):
        """Push changes to the default Git repository remote."""
        self._git("push")

    def get_commits(self, limit: int = None, skip: int = None) -> List[Commit]:
        """
        Return commits for the current branch. By default, all will be returned
        as a list. Alternatively a limit (and optionall a skip offset) may be
        specified to reduce the number of commits returned.
        """
        args = ["log", "--format=h:%t%nm:%s%nd:%aI%na:%an%ne:%aE%nc:%cn%n--"]
        if limit:
            args.append(f"--max-count={limit}")
        if skip:
            args.append(f"--skip={skip}")

        proc = self._git(*args)
        raw_commits = proc.stdout.split("\n--")
        commits: List[Commit] = []
        for raw_commit in raw_commits:
            commit_hash = message = date = author_name = author_email = (
                committer_name
            ) = None
            lines = raw_commit.strip().splitlines()
            if not len(lines) == 6:
                continue
            for line in lines:
                key: str = line[0]
                val: str = line[2:]
                match key:
                    case "h":
                        commit_hash = val.strip()
                    case "m":
                        message = val.strip()
                    case "d":
                        date = datetime.fromisoformat(val.strip())
                    case "a":
                        author_name = val.strip()
                    case "e":
                        author_email = val.strip()
                    case "c":
                        committer_name = val.strip()
            commit = Commit(
                commit_hash, message, date, author_name, author_email, committer_name
            )
            commits.append(commit)
        return commits


def has_git_repo_subdirs(path: str | Path) -> bool:
    """Determine if `path` contains subdirectories that are Git repositories."""
    path = Path(path).resolve()
    if not path.is_dir():
        raise NotADirectoryError(f"{path} is not a directory")
    for sub_path in path.iterdir():
        if sub_path.is_dir():
            if GitRepo.is_repository(sub_path):
                return True
    return False


def get_git_dirs(path: str | Path) -> List[Path]:
    """Return all subdirectories  of `path` that contain Git repositories."""
    path = Path(path).resolve()
    if not path.is_dir():
        raise NotADirectoryError(f"{path} is not a directory")
    git_dirs = []
    for sub_path in path.iterdir():
        if sub_path.is_dir():
            if GitRepo.is_repository(sub_path):
                git_dirs.append(sub_path)
    return git_dirs


def get_git_repos(path: str | Path) -> List[GitRepo]:
    """Return all git repositories that are subdirectories of `path`."""
    path = Path(path).resolve()
    if not path.is_dir():
        raise NotADirectoryError(f"{path} is not a directory")
    repos = []
    for sub_path in path.iterdir():
        if sub_path.is_dir():
            try:
                repos.append(GitRepo(sub_path))
            except NotARepoError:
                pass
    return repos


def get_dirty_repos(path: str | Path, untracked_files: bool = True) -> List[GitRepo]:
    """
    Return a list of GitRepo objects for all subdirectories of `path` if and
    only if the Git repository has changed or untracked files.
    """
    repos = get_git_repos(path)
    return [repo for repo in repos if repo.is_dirty(untracked_files=untracked_files)]



import click
import sys
import functools
import logging

from au.classroom import (
    AssignmentSettings,
    Assignment,
    choose_assignment,
    choose_classroom,
    get_assignment,
)

logger = logging.getLogger(__name__)


class AssignmentOptions:
    def __init__(
        self,
        load: bool = True,
        store: bool = True,
        force_store: bool = False,
        required: bool = False,
    ):
        self.load = load
        self.store = store
        self.force_store = force_store
        self.required = required

    def get_settings(self):
        ctx = click.get_current_context()
        base_path = ctx.meta.get("au.base_path")  # Command must use BasePath type
        if base_path and (self.load or self.store):
            try:
                return AssignmentSettings.get_classroom_settings(base_path)
            except:
                if self.store:
                    if self.force_store or AssignmentSettings.is_valid_settings_path(
                        base_path
                    ):
                        return AssignmentSettings(base_path, create=True)
        return None

    def options(self, func):
        help_text = "The integer classroom id for the assignment."
        if self.load:
            help_text += (
                " If not provided, will try to read a value stored in settings."
            )
        help_text += " Will prompt for for an assignment interactively if not provided."
        if self.required:
            help_text += " An assignment is REQUIRED to run this command."

        @click.option("--assignment-id", type=int, help=help_text)
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            kwargs["assignment"] = None  # default to None, but set below
            settings = self.get_settings()
            assignment_id: int = None
            if "assignment_id" in kwargs:
                assignment_id = kwargs.pop("assignment_id")
            if self.load and not assignment_id:
                if settings and settings.assignment_id:
                    assignment_id = settings.assignment_id
            if assignment_id:
                assignment = get_assignment(assignment_id)
            else:
                assignment: Assignment = None
                classroom = choose_classroom()
                if classroom:
                    assignment = choose_assignment(classroom)
                elif self.required:
                    click.echo("An assignment is required for this operation.\n")
                    click.echo(click.get_current_context().get_help())
                    sys.exit(1)
            if assignment:
                kwargs["assignment"] = assignment
                try:
                    if settings and self.store:
                        with settings:
                            settings.classroom_id = assignment.classroom.id
                            settings.assignment_id = assignment.id
                except Exception:
                    logger.exception("Error encountered while writing settings")
            elif assignment_id:
                logger.error(f"Unable to find assignment with id = {assignment_id}")
                sys.exit(1)
            elif self.required:
                click.echo("An assignment is required for this operation.\n")
                click.echo(click.get_current_context().get_help())
                sys.exit(1)
            return func(*args, **kwargs)

        return wrapper



from typing import Dict

import csv as csv_lib
from io import TextIOBase, StringIO
from pathlib import Path

import logging

logger = logging.getLogger(__name__)


def dict_from_csv(
    csv: Path | str | TextIOBase,
    key_col: int | str,
    val_col: int | str,
    no_header: bool = False,
) -> Dict[any, any]:
    """
    Convert a CSV formatted text into a dictionary using the `key_col` column in
    the CSV as key and `val_col` for values. If duplicate key's are found in the
    CSV, the value will be derived from the last one processed. All keys and
    values will be of type str.
    """

    csv_dict = {}
    open_file = None

    try:
        logger.debug(f"Checking {csv}")
        if isinstance(csv, str):
            logger.debug(f"Found str")
            csv_iter = StringIO(csv)
        elif isinstance(csv, Path):
            logger.debug(f"Found Path")
            if not csv.exists():
                raise FileNotFoundError()
            logger.debug(f"Opening {csv}")
            open_file = open(csv)
            csv_iter = open_file
        elif isinstance(csv, TextIOBase):
            logger.debug(f"Found TextIOBase")
            csv_iter = csv
        else:
            raise ValueError("csv is not a string, stream, or Path")

        if isinstance(key_col, int) and isinstance(val_col, int):
            key_idx = key_col
            val_idx = val_col
        elif isinstance(key_col, str) and isinstance(val_col, str):
            if no_header:
                raise ValueError(
                    "no_header cannot be True if key_col and val_col are str"
                )
            key_idx = None
            val_idx = None
        else:
            raise ValueError("key_col and val_col must both either be int or str")

        logger.debug(f"Using key_idx={key_idx} and val_idx={val_idx}")

        rdr = csv_lib.reader(csv_iter)

        logger.debug(f"{rdr}")

        header_skipped = no_header

        for row in rdr:
            logger.debug(f"PROCESSING: {row}")
            if header_skipped:
                key = row[key_idx].strip()
                val = row[val_idx].strip()
                if key:
                    csv_dict[key] = val
            else:
                if not key_idx:
                    # Read header to match column to key_idx and val_idx.
                    # Allows unhandled ValueError to be raised as this is fatal.
                    row_lower = [text.lower() for text in row]
                    key_idx = row_lower.index(key_col.lower())
                    val_idx = row_lower.index(val_col.lower())
                header_skipped = True
    except:
        raise
    finally:
        if open_file:
            open_file.close()
        return csv_dict



from enum import Enum


class BoxChars(Enum):
    """Unicode box drawing characters"""

    LIGHT_HORIZONTAL = "\u2500"
    HEAVY_HORIZONTAL = "\u2501"
    LIGHT_VERTICAL = "\u2502"
    HEAVY_VERTICAL = "\u2503"
    LIGHT_DOWN_AND_RIGHT = "\u250c"
    HEAVY_DOWN_AND_RIGHT = "\u250f"
    LIGHT_DOWN_AND_LEFT = "\u2510"
    HEAVY_DOWN_AND_LEFT = "\u2513"
    LIGHT_UP_AND_RIGHT = "\u2514"
    HEAVY_UP_AND_RIGHT = "\u2517"
    LIGHT_UP_AND_LEFT = "\u2518"
    HEAVY_UP_AND_LEFT = "\u251b"
    LIGHT_VERTICAL_AND_RIGHT = "\u251c"
    HEAVY_VERTICAL_AND_RIGHT = "\u2523"
    LIGHT_VERTICAL_AND_LEFT = "\u2524"
    HEAVY_VERTICAL_AND_LEFT = "\u252b"
    LIGHT_DOWN_AND_HORIZONTAL = "\u252c"
    HEAVY_DOWN_AND_HORIZONTAL = "\u2533"
    LIGHT_UP_AND_HORIZONTAL = "\u2534"
    HEAVY_UP_AND_HORIZONTAL = "\u253b"
    LIGHT_VERTICAL_AND_HORIZONTAL = "\u253c"
    HEAVY_VERTICAL_AND_HORIZONTAL = "\u254b"
    LIGHT_DOUBLE_DASH_HORIZONTAL = "\u254c"
    HEAVY_DOUBLE_DASH_HORIZONTAL = "\u254d"
    LIGHT_DOUBLE_DASH_VERTICAL = "\u254e"
    HEAVY_DOUBLE_DASH_VERTICAL = "\u254f"
    DOUBLE_HORIZONTAL = "\u2550"
    DOUBLE_VERTICAL = "\u2551"
    DOUBLE_DOWN_AND_RIGHT = "\u2554"
    DOUBLE_DOWN_AND_LEFT = "\u2557"
    DOUBLE_UP_AND_RIGHT = "\u255a"
    DOUBLE_UP_AND_LEFT = "\u255d"
    DOUBLE_VERTICAL_AND_RIGHT = "\u2560"
    DOUBLE_VERTICAL_AND_LEFT = "\u2563"
    DOUBLE_DOWN_AND_HORIZONTAL = "\u2566"
    DOUBLE_UP_AND_HORIZONTAL = "\u2569"
    DOUBLE_VERTICAL_AND_HORIZONTAL = "\u256c"
    DOUBLE_VERTICAL_AND_HEAVY_RIGHT = "\u2561"
    DOUBLE_VERTICAL_AND_HEAVY_LEFT = "\u2562"
    DOUBLE_DOWN_AND_HEAVY_HORIZONTAL = "\u2564"
    DOUBLE_UP_AND_HEAVY_HORIZONTAL = "\u2567"
    DOUBLE_VERTICAL_AND_HEAVY_HORIZONTAL = "\u256a"
    LIGHT_VERTICAL_AND_DOUBLE_RIGHT = "\u255f"
    LIGHT_VERTICAL_AND_DOUBLE_LEFT = "\u2562"
    LIGHT_DOWN_AND_DOUBLE_HORIZONTAL = "\u2565"
    LIGHT_UP_AND_DOUBLE_HORIZONTAL = "\u2568"
    LIGHT_VERTICAL_AND_DOUBLE_HORIZONTAL = "\u256b"
    LIGHT_ARC_DOWN_AND_RIGHT = "\u256d"
    LIGHT_ARC_DOWN_AND_LEFT = "\u256e"
    LIGHT_ARC_UP_AND_LEFT = "\u256f"
    LIGHT_ARC_UP_AND_RIGHT = "\u2570"
    HEAVY_ARC_DOWN_AND_RIGHT = "\u2571"
    HEAVY_ARC_DOWN_AND_LEFT = "\u2572"
    HEAVY_ARC_UP_AND_LEFT = "\u2573"
    HEAVY_ARC_UP_AND_RIGHT = "\u2574"

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value


class TableStyle:
    def __init__(self):
        self.top_border = True
        self.bottom_border = True
        self.terminal_style = True
        self.allow_lazy_header = True
        self.force_header = False
        self.align_char = None

        self.cell_padding = 1
        self.min_width = 1

        self.no_header_top_line = BoxChars.LIGHT_HORIZONTAL
        self.no_header_top_delimiter = BoxChars.LIGHT_DOWN_AND_HORIZONTAL
        self.no_header_top_left = BoxChars.LIGHT_DOWN_AND_RIGHT
        self.no_header_top_right = BoxChars.LIGHT_DOWN_AND_LEFT

        self.header_top_line = BoxChars.LIGHT_HORIZONTAL
        self.header_top_delimiter = BoxChars.LIGHT_DOWN_AND_HORIZONTAL
        self.header_top_left = BoxChars.LIGHT_DOWN_AND_RIGHT
        self.header_top_right = BoxChars.LIGHT_DOWN_AND_LEFT

        self.header_delimiter = BoxChars.LIGHT_VERTICAL
        self.header_left = BoxChars.LIGHT_VERTICAL
        self.header_right = BoxChars.LIGHT_VERTICAL

        self.header_bottom_line = BoxChars.LIGHT_HORIZONTAL
        self.header_bottom_delimiter = BoxChars.LIGHT_VERTICAL_AND_HORIZONTAL
        self.header_bottom_left = BoxChars.LIGHT_VERTICAL_AND_RIGHT
        self.header_bottom_right = BoxChars.LIGHT_VERTICAL_AND_LEFT

        self.values_delimiter = BoxChars.LIGHT_VERTICAL
        self.values_left = BoxChars.LIGHT_VERTICAL
        self.values_right = BoxChars.LIGHT_VERTICAL

        self.values_bottom_line = BoxChars.LIGHT_HORIZONTAL
        self.values_bottom_delimiter = BoxChars.LIGHT_UP_AND_HORIZONTAL
        self.values_bottom_left = BoxChars.LIGHT_UP_AND_RIGHT
        self.values_bottom_right = BoxChars.LIGHT_UP_AND_LEFT

        self.row_separator_line = "◦"
        self.row_separator_delimiter = BoxChars.LIGHT_VERTICAL
        self.row_separator_left = BoxChars.LIGHT_VERTICAL
        self.row_separator_right = BoxChars.LIGHT_VERTICAL


class BasicScreenStyle(TableStyle): ...


class RoundedBorderScreenStyle(TableStyle):
    def __init__(self):
        super().__init__()
        self.header_top_left = BoxChars.LIGHT_ARC_DOWN_AND_RIGHT
        self.header_top_right = BoxChars.LIGHT_ARC_DOWN_AND_LEFT
        self.no_header_top_left = BoxChars.LIGHT_ARC_DOWN_AND_RIGHT
        self.no_header_top_right = BoxChars.LIGHT_ARC_DOWN_AND_LEFT
        self.values_bottom_left = BoxChars.LIGHT_ARC_UP_AND_RIGHT
        self.values_bottom_right = BoxChars.LIGHT_ARC_UP_AND_LEFT


class NoBorderScreenStyle(TableStyle):
    def __init__(self):
        super().__init__()
        self.top_border = False
        self.bottom_border = False
        self.header_left = ""
        self.header_right = ""
        self.header_bottom_left = ""
        self.header_bottom_right = ""
        self.values_left = ""
        self.values_right = ""
        self.values_bottom_left = ""
        self.values_bottom_right = ""
        self.row_separator_line = "◦"
        self.row_separator_left = ""
        self.row_separator_right = ""


class MarkdownStyle(TableStyle):
    def __init__(self):
        super().__init__()
        self.top_border = False
        self.bottom_border = False
        self.terminal_style = False
        self.allow_lazy_header = False
        self.force_header = True
        self.align_char = ":"

        self.min_width = 3

        self.header_delimiter = "|"
        self.header_left = "|"
        self.header_right = "|"

        self.header_bottom_line = "-"
        self.header_bottom_delimiter = "|"
        self.header_bottom_left = "|"
        self.header_bottom_right = "|"

        self.values_delimiter = "|"
        self.values_left = "|"
        self.values_right = "|"

        self.row_separator_line = ""
        self.row_separator_delimiter = ""
        self.row_separator_left = ""
        self.row_separator_right = ""



from datetime import datetime, timezone, timedelta


def parse_github_datetime(datetime_str: str) -> datetime:
    try:
        return datetime.strptime(datetime_str, "%Y-%m-%dT%H:%M:%SZ").replace(
            tzinfo=timezone.utc
        )
    except:
        return None


def utc_now():
    return datetime.now(timezone.utc)


def local_now():
    return date_to_local(utc_now())


def utc_min():
    return datetime.min.replace(tzinfo=timezone.utc)


def utc_max():
    return datetime.max.replace(tzinfo=timezone.utc)


def format_github_datetime(dt: datetime) -> str:
    return datetime.strftime(dt, "%Y-%m-%dT%H:%M:%SZ")


def date_to_utc(to_convert: datetime) -> datetime:
    return to_convert.astimezone(timezone.utc)


def date_to_local(to_convert: datetime) -> datetime:
    return to_convert.astimezone()


def get_friendly_local_datetime(dt: datetime) -> str:
    try:
        return date_to_local(dt).strftime("%Y-%m-%d %I:%M %p")
    except:
        return "N/A"


def get_friendly_timedelta(delta: timedelta) -> str:
    pd_str = ""
    if delta.days:
        pd_str += f"{delta.days} day + "
    secs = delta.seconds
    hours = secs // 3600
    secs = secs % 3600
    mins = secs // 60
    pd_str += f"{hours}:{mins:02d}"
    return pd_str



from typing import Iterable
from os import get_terminal_size, sep
from math import ceil
import re
from pathlib import Path
from beaupy import select


from .drawing import get_line, BoxChars

###############################################################################
# get_term_width
###############################################################################

MAX_REASONABLE_WIDTH = 120


def get_term_width(max_term_width: int = MAX_REASONABLE_WIDTH):
    try:
        term_width = get_terminal_size().columns
    except:
        return max_term_width
    if max_term_width == 0:
        max_term_width = 9999
    return min(max_term_width, term_width)


###############################################################################
# get_line / draw_line / draw_single_line / draw_double_line
###############################################################################


def draw_single_line(text: str = "", max_width: int = MAX_REASONABLE_WIDTH) -> None:
    draw_line(BoxChars.HEAVY_HORIZONTAL, text, max_width)


def draw_double_line(text: str = "", max_width: int = MAX_REASONABLE_WIDTH) -> None:
    draw_line(BoxChars.DOUBLE_HORIZONTAL, text, max_width)


def draw_line(
    char: str = BoxChars.HEAVY_HORIZONTAL,
    text: str = "",
    max_width: int = MAX_REASONABLE_WIDTH,
) -> None:
    term_width = get_term_width(max_width)
    print(get_line(char, text, term_width))


###############################################################################
# get_choice
###############################################################################


def get_choice(
    choices: Iterable,
    title: str = None,
    prompt: str = "Choose",
    ideal_max_rows: int = 8,
    max_term_width: int = MAX_REASONABLE_WIDTH,
) -> int:
    """
    Allow a user to choose from a list of choices. Display the choices in
    columns if the number of choices exceeds ideal_max_rows.
    """
    term_width = get_term_width(max_term_width)
    ideal_col_count = ceil(len(choices) / ideal_max_rows)
    max_len = max([len(choice) for choice in choices])
    col_len = max_len + 6
    max_col_count = term_width // col_len
    col_count = min(ideal_col_count, max_col_count)
    row_count = ceil(len(choices) / col_count)

    if title:
        draw_double_line(max_width=term_width)
        print(title)
        draw_single_line(max_width=term_width)

    for i in range(row_count):
        for j in range(col_count):
            idx = i + j * row_count
            if idx < len(choices):
                choice_num = idx + 1
                print(f"{choice_num:>2}. {choices[idx]}".ljust(col_len), end="")
        print()
    draw_single_line(max_width=term_width)

    while True:
        try:
            choice = int(input(f"{prompt} >> "))
            assert 0 < choice <= len(choices)
            return choice - 1
        except (AssertionError, ValueError):
            print("!! INVALID CHOICE !!")


###############################################################################
# select_choice
###############################################################################


def select_choice(
    choices: Iterable,
    title: str = None,
    max_rows: int = 8,
    max_term_width: int = MAX_REASONABLE_WIDTH,
) -> int | None:
    term_width = get_term_width(max_term_width)

    if title:
        draw_double_line(max_width=term_width)
        print(title)
        draw_single_line(max_width=term_width)
    return select(
        options=choices, pagination=True, page_size=max_rows, return_index=True
    )


###############################################################################
# select_file
###############################################################################


def select_file(
    root: Path = None,
    filter: str | Iterable[str] = "*",
    title: str = None,
    files_at_top: bool = True,
    max_rows: int = 8,
    max_term_width: int = MAX_REASONABLE_WIDTH,
) -> Path | None:
    """Interactive file picker."""
    if not root:
        root = Path.cwd().resolve()
    term_width = get_term_width(max_term_width)

    if title:
        draw_double_line(max_width=term_width)
        print(title)
        draw_single_line(max_width=term_width)

    path = root
    while True:
        options = []
        options.append(f".{sep}\t\t({path})")
        sel_start = 1
        if path.parent and path.parent != path:
            options.append(f"..{sep}\t\t({path.parent})")
            sel_start = 2
        if isinstance(filter, str):
            files = [file.name for file in path.glob(filter) if file.is_file()]
        else:
            files = []
            for pattern in filter:
                files += [file.name for file in path.glob(pattern) if file.is_file()]
        dirs = [file.name + sep for file in path.iterdir() if file.is_dir()]
        files.sort(key=str.casefold)
        dirs.sort(key=str.casefold)
        all = files + dirs
        if not all:
            sel_start -= 1
        if not files_at_top:
            all.sort(key=str.casefold)
        options.extend(all)

        if not max_rows or max_rows < 3:
            pagination = False
            max_rows = 1  # beaupy blows up if < 1
        else:
            pagination = True

        sel = select(
            options=options,
            cursor_index=sel_start,
            pagination=pagination,
            page_size=max_rows,
        )

        if not sel:
            path = None
            break
        sel = sel.split(sep)[0]
        path = (path / sel).resolve()
        if path.is_file():
            break
    return path


###############################################################################
# select_dir
###############################################################################


def select_dir(
    root: Path = None,
    title: str = None,
    exclude_hidden: bool = True,
    exclude_dunder: bool = True,
    max_rows: int = 8,
    max_term_width: int = MAX_REASONABLE_WIDTH,
) -> Path | None:
    """Interactive directory picker."""
    if not root:
        root = Path.cwd().resolve()
    term_width = get_term_width(max_term_width)

    if title:
        draw_double_line(max_width=term_width)
        print(title)
        draw_single_line(max_width=term_width)

    path = root
    while True:
        options = []
        options.append(f"CHOOSE {path}")
        if path.parent and path.parent != path:
            options.append(f"..{sep}")
        dirs = []
        for file in path.iterdir():
            if not file.is_dir():
                continue
            if exclude_hidden and file.name.startswith("."):
                continue
            if exclude_dunder and file.name.startswith("__"):
                continue
            dirs.append(file.name + sep)
        dirs.sort(key=str.casefold)
        options.extend(dirs)

        if not max_rows or max_rows < 3:
            pagination = False
            max_rows = 1  # beaupy blows up if < 1
        else:
            pagination = True

        sel = select(
            options=options,
            pagination=pagination,
            page_size=max_rows,
            return_index=True,
        )

        if sel == None:
            path = None
            break
        if sel == 0:
            return path
        path = (path / options[sel]).resolve()
    return path


###############################################################################
# clean_ansi
###############################################################################

ANSI_ESCAPE_PATTERN = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")


def strip_ansi(text: str):
    return ANSI_ESCAPE_PATTERN.sub("", text)



import pytest
from pytest import TestReport

from .pytest_data import Results, NodeId, TestClass, Test, SubTest


class PytestResultsReporter:
    """
    Custom results reporter that formats results as JSON suitable for reporting
    to students.
    """

    def __init__(self):
        self.results = Results()
        self.last_err = None
        self.config = None

    # def pytest_report_teststatus(self, report: TestReport):
    #     '''
    #     Basically just want to eliminate the 1 character results printing
    #     '''
    #     category, short, verbose = '', '', ''
    #     if hasattr(report, 'wasxfail'):
    #         if report.skipped:
    #             category = 'xfailed'
    #             verbose = 'xfail'
    #         elif report.passed:
    #             category = 'xpassed'
    #             verbose = ('XPASS', {'yellow': True})
    #         return (category, short, verbose)
    #     elif report.when in ('setup', 'teardown'):
    #         if report.failed:
    #             category = 'error'
    #             verbose = 'ERROR'
    #         elif report.skipped:
    #             category = 'skipped'
    #             verbose = 'SKIPPED'
    #         return (category, short, verbose)
    #     category = report.outcome
    #     verbose = category.upper()
    #     return (category, short, verbose)

    def pytest_runtest_logreport(self, report: TestReport):
        """
        Process a test setup / call / teardown report.
        """

        # ignore successful setup and teardown stages
        if report.passed and report.when != "call":
            return

        test = self.results.get_test(report.nodeid)
        state = test
        sub_name = SubTest.parse_head_line(report.head_line)
        if sub_name:
            sub_test = test.get_subtest(sub_name)
            state = sub_test

        # Store duration
        state.duration = report.duration

        # Update tests that have already failed with capstdout and return.
        if not state.is_passing():
            if report.capstdout.rstrip("FFFFFFFF ").rstrip("uuuuu"):
                state.output = report.capstdout.rstrip("FFFFFFFF ").rstrip("uuuuu")
            return

        # Record captured relevant stdout content for passed tests.
        if report.capstdout:
            state.output = report.capstdout

        # Handle details of test failure
        if report.failed:

            # traceback that caused the issued, if any
            message = None
            if report.longrepr:
                trace = report.longrepr.reprtraceback
                crash = report.longrepr.reprcrash
                message = self._make_message(trace, crash)

            # test failed due to a setup / teardown error
            if report.when != "call":
                state.error(message)
            else:
                state.fail(message)

    def pytest_sessionfinish(self, session, exitstatus):
        """Processes the results into a report."""
        exitcode = pytest.ExitCode(int(exitstatus))

        # at least one of the tests has failed
        if (
            exitcode is not pytest.ExitCode.TESTS_FAILED
            and exitcode is not pytest.ExitCode.OK
        ):
            message = None
            if self.last_err is not None:
                message = self.last_err
            else:
                message = f"Unexpected ExitCode.{exitcode.name}: check logs for details"
            self.results.error(message)
        self.results.update()

    def pytest_exception_interact(self, node, call, report):
        """Catch the last exception handled in case the test run itself errors."""
        if report.outcome == "failed":
            excinfo = call.excinfo
            err = excinfo.getrepr(style="no", abspath=False)

            trace = err.chain[-1][0]
            crash = err.chain[0][1]
            self.last_err = self._make_message(trace, crash)

    def _make_message(self, trace, crash):
        """Make a formatted message for reporting."""
        if crash:
            message = ""
            if "<string>" in crash.path:
                message = f"Error at line {crash.lineno}:\n"
            elif "unittest/case.py" not in crash.path:
                message = f"Error in {crash.path} at line {crash.lineno}:\n"
            message += crash.message
        else:
            message = "\n".join(trace.reprentries[-1].lines)
        return message



import logging
from os.path import commonprefix as get_common_prefix
from pathlib import Path
from pprint import pformat

import click

from au.classroom import Roster
from au.click import RosterOptions, DebugOptions, BasePath


logger = logging.getLogger(__name__)


@click.command()
@click.argument("root_dir", type=BasePath(), default=".")
@RosterOptions(prompt=True, required=True).options
@click.option(
    "--preserve-prefix",
    is_flag=True,
    help="set to preserve the prefix (slug) string common to all repositories",
)
@click.option(
    "-p",
    "--preview",
    is_flag=True,
    help="set to show changes without actually making them",
)
@DebugOptions().options
def rename_roster(
    root_dir: Path,
    roster: Roster,
    preserve_prefix: bool = False,
    preview: bool = False,
    **kwargs,
) -> None:
    """Rename subdirectories to contain students' real names.

    This is generally only useful if student repositories have already been
    cloned using other means. This command uses the same naming logic as does
    `au classroom clone-all`.

    This will rename the subdirectories in ROOT_DIR to contain a students real
    name by matching a the Github ID from the classroom roster to a folder name.
    Any potentially unsafe characters, as well as commas and spaces, are
    replaced with an underscore (`_`). Any characters after the matching github
    id are preserved in order to prevent possible duplicate directory names.

    The purpose is to help with finding and sorting directories by the real
    names of students.

    The resulting name will be:

        [real name]@[github id][suffix]/

    \For example:

        York_Paul@ptyork/

    Directories of students that are not in the roster are skipped entirely. If
    the student's name is found in the directory name, it is likewise skipped as
    it is assumed that the directory has already been renamed.

    If ROOT_DIR is not provided, then the current working directory will be
    assumed.
    """
    logging.basicConfig()

    logger.debug(pformat(root_dir))
    logger.debug(pformat(roster))

    login_dir_map = roster.get_login_dir_map(root_dir)
    common_prefix = get_common_prefix(list(login_dir_map.values()))
    prefix = ""
    if preserve_prefix:
        prefix = common_prefix
    login_dirname_map = roster.get_dir_names(prefix)

    logger.debug(f"Renaming subdirectories in {root_dir}")
    for login, old_dir_name in login_dir_map.items():
        new_dir_name = login_dirname_map.get(login)
        if old_dir_name and new_dir_name and old_dir_name != new_dir_name:
            old_dir = root_dir / old_dir_name
            if preview:
                logger.info(f"Would rename {old_dir_name} to {new_dir_name}")
            else:
                logger.info(f"Renaming {old_dir_name} to {new_dir_name}")
                try:
                    old_dir.rename(new_dir_name)
                except Exception as ex:
                    logger.error(f'Unable to rename "{old_dir}": {ex}')
        else:
            logger.info(f"SKIPPING: {old_dir_name}")


if __name__ == "__main__":
    rename_roster()



import click
from rich.console import Console

from f_table import get_table, BasicScreenStyle

from au.click import AliasedGroup, AssignmentOptions, RosterOptions, BasePath
from au.classroom import (
    AssignmentSettings,
    Roster,
    get_classroom,
    choose_classroom,
    get_accepted_assignments,
    AcceptedAssignment,
)

from .rename_roster import rename_roster
from .commit_all import commit_all
from .clone_all import clone_all_cmd
from .time_details import time_details


@click.group(cls=AliasedGroup)
def assignment():
    """Commands for working with any GitHub Classroom Assignment."""


assignment.add_command(rename_roster)
assignment.add_command(commit_all)
assignment.add_command(clone_all_cmd)
assignment.add_command(time_details)


@assignment.command()
@click.argument("assignment_dir", type=BasePath(), default=".")
@AssignmentOptions(required=True, load=False, store=True, force_store=True).options
@RosterOptions(load=False, store=True, prompt=True).options
def config(assignment_dir, **kwargs):
    """Create or change settings for ASSIGNMENT_DIR.

    Most commands that accept configuration parameters such as an assignment-id
    or roster file will save this to a local configuration file. However, this
    command is useful if you need to change the settings or if you are
    pre-configuring an assignment directory prior to cloning student
    submissions.

    If not specified, ASSIGNMENT_DIR defaults to the current directory.
    """
    settings = AssignmentSettings(assignment_dir)
    if settings:
        print(f"Settings saved in {settings.settings_doc_path / settings.FILENAME}")
    else:
        print(f"Error encountered while configuring {assignment_dir}")


@assignment.command()
@AssignmentOptions(required=True, store=False).options
def info(assignment):
    """Display details for an assignment.

    Details include the due date, assignment type, and a count of accepted and
    submitted assignments, among others.

    If executed inside of a configured assignment directory, it will display the
    information for the configured assignment. If not, it will prompt the user
    to select a classroom and assignment.
    """
    print(assignment.as_table())


@assignment.command()
@AssignmentOptions(required=True, store=False).options
@RosterOptions(required=False, load=False, store=False, prompt=True).options
def accepted(assignment, roster: Roster):
    """List all students that have accepted an  assignment."""
    with Console().status(
        "Retrieving data from GitHub Classroom", spinner="bouncingBall"
    ):
        rows = []
        for assn in get_accepted_assignments(assignment):
            name = None
            if assn.login:
                if roster:
                    name = roster.get_name(assn.login)
                if not name:
                    name = assn.login
            if not name:
                name = "NO NAME"
            rows.append([name, assn.commit_count, assn.repository.html_url])
        rows.sort(key=lambda row: row[0].casefold())
    print(
        get_table(
            header_row=["STUDENT", "COMMITS", "REPOSITORY"],
            value_rows=rows,
            col_defs=["", "^", "A"],
            style=BasicScreenStyle(),
        )
    )


if __name__ == "__main__":
    assignment()
